from __future__ import print_function
"""Search a FingerprintArena and work with the search results

This module implements the different ways to search a
:class:`FingerprintArena`. The search functions are:

  Count the number of hits:
  * count_tanimoto_hits_fp - search an arena using a single fingerprint
  * count_tanimoto_hits_arena - search an arena using an arena
  * count_tanimoto_hits_symmetric - search an arena using itself
  * partial_count_tanimoto_hits_symmetric - (advanced use; see the doc string)

  Find all hits at or above a given threshold, sorted arbitrarily:
  * threshold_tanimoto_search_fp - search an arena using a single fingerprint
  * threshold_tanimoto_search_arena - search an arena using an arena
  * threshold_tanimoto_search_symmetric - search an arena using itself
  * partial_threshold_tanimoto_search_symmetric - (advanced use; see the doc string)
  * fill_lower_triangle - copy the upper triangle terms to the lower triangle

  Find the k-nearest hits at or above a given threshold, sorted by
  decreasing similarity:
  * knearest_tanimoto_search_fp - search an arena using a single fingerprint
  * knearest_tanimoto_search_arena - search an arena using an arena
  * knearest_tanimoto_search_symmetric - search an arena using itself

The threshold and k-nearest search results use a :class:`SearchResult` when
a fingerprint is used as a query, or a :class:`SearchResults` when an arena
is used as a query. These internally use a compressed sparse row format.
"""
# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

import _chemfp
import ctypes
import array

# 
__all__ = ["SearchResult", "SearchResults",
           "count_tanimoto_hits_fp", "count_tanimoto_hits_arena",
           "count_tanimoto_hits_symmetric", "partial_count_tanimoto_hits_symmetric",

           "threshold_tanimoto_search_fp", "threshold_tanimoto_search_arena",
           "threshold_tanimoto_search_symmetric", "partial_threshold_tanimoto_search_symmetric",
           "fill_lower_triangle",

           "knearest_tanimoto_search_fp", "knearest_tanimoto_search_arena",
           "knearest_tanimoto_search_symmetric"
           ]
           

class SearchResult(object):
    """Search results for a query fingerprint against a target arena.

    The results contains a list of hits. Hits contain a target index,
    score, and optional target ids. The hits can be reordered based on
    score or index.
    
    """
    def __init__(self, search_results, row):
        "The constructor is not part of the public API"
        self._search_results = search_results
        self._row = row

    def __len__(self):
        """The number of hits"""
        return self._search_results._size(self._row)

    def __iter__(self):
        """Iterate through the pairs of (target index, score) using the current ordering"""
        return iter(self._search_results._get_indices_and_scores(self._row))
        
    def clear(self):
        """Remove all hits from this result"""
        self._search_results._clear_row(self._row)

    def get_indices(self):
        """The list of target indices, in the current ordering."""
        return self._search_results._get_indices(self._row)

    def get_ids(self):
        """The list of target identifiers (if available), in the current ordering"""
        ids = self._search_results.target_ids
        if ids is None:
            return None
        return [ids[i] for i in self._search_results._get_indices(self._row)]

    def iter_ids(self):
        """Iterate over target identifiers (if available), in the current ordering"""
        ids = self._search_results.target_ids
        if ids is None:
            return
        return (ids[i] for i in self._search_results._get_indices(self._row))
    
    def get_scores(self):
        """The list of target scores, in the current ordering"""
        return self._search_results._get_scores(self._row)

        
    def get_ids_and_scores(self):
        """The list of (target identifier, target score) pairs, in the current ordering

        Raises a TypeError if the target IDs are not available.
        """
        ids = self._search_results.target_ids
        if ids is None:
            raise TypeError("target_ids are not available")
        return zip(self.get_ids(), self.get_scores())

    def get_indices_and_scores(self):
        """The list of (target index, score) pairs, in the current ordering"""
        return self._search_results._get_indices_and_scores(self._row)
            
    def reorder(self, ordering="decreasing-score"):
        """Reorder the hits based on the requested ordering.

        The available orderings are:
          * increasing-score - sort by increasing score
          * decreasing-score - sort by decreasing score
          * increasing-index - sort by increasing target index
          * decreasing-index - sort by decreasing target index
          * move-closest-first - move the hit with the highest score to the first position
          * reverse - reverse the current ordering

        :param string ordering: the name of the ordering to use
        """
        self._search_results._reorder_row(self._row, ordering)

    def count(self, min_score=None, max_score=None, interval="[]"):
        """Count the number of hits with a score between *min_score* and *max_score*

        Using the default parameters this returns the number of
        hits in the result.

        The default *min_score* of None is equivalent to -infinity.
        The default *max_score* of None is equivalent to +infinity.
        
        The *interval* parameter describes the interval end
        conditions. The default of "[]" uses a closed interval,
        where min_score <= score <= max_score. The interval "()"
        uses the open interval where min_score < score < max_score.
        The half-open/half-closed intervals "(]" and "[)" are
        also supported.

        :param min_score: the minimum score in the range.
        :type min_score: a float, or None for -infinity
        :param max_score: the maximum score in the range.
        :type max_score: a float, or None for +infinity
        :param interval: specify if the end points are open or closed.
        :type interval: one of "[]", "()", "(]", "[)"
        :returns: an integer count
        """
        return self._search_results._count_row(self._row, min_score, max_score, interval)

    def cumulative_score(self, min_score=None, max_score=None, interval="[]"):
        """The sum of the scores which are between *min_score* and *max_score*

        Using the default parameters this returns the sum of all of
        the scores in the result. With a specified range this returns
        the sum of all of the scores in that range. The cumulative
        score is also known as the raw score.
        
        The default *min_score* of None is equivalent to -infinity.
        The default *max_score* of None is equivalent to +infinity.
        
        The *interval* parameter describes the interval end
        conditions. The default of "[]" uses a closed interval,
        where min_score <= score <= max_score. The interval "()"
        uses the open interval where min_score < score < max_score.
        The half-open/half-closed intervals "(]" and "[)" are
        also supported.
        
        :param min_score: the minimum score in the range.
        :type min_score: a float, or None for -infinity
        :param max_score: the maximum score in the range.
        :type max_score: a float, or None for +infinity
        :param interval: specify if the end points are open or closed.
        :type interval: one of "[]", "()", "(]", "[)"
        :returns: a floating point value
        """
        return self._search_results._cumulative_score_row(self._row, min_score, max_score, interval)

    ## ??? What does this do?
    ## @property
    ## def target_id(self):
    ##     ids = self._search_results.target_ids
    ##     if ids is None:
    ##         return None
    ##     return ids[self._row]

class SearchResults(_chemfp.SearchResults):
    """Search results for a list of query fingerprints against a target arena

    This acts like a list of SearchResult elements, with the ability
    to iterate over each search results, look them up by index, and
    get the number of scores.

    In addition, there are helper methods to iterate over each hit and
    to get the hit indicies, scores, and identifiers directly as Python
    lists, sort the list contents, and more.
    
    """
    def __init__(self, num_rows, num_cols, arena_ids=None):
        """*num_rows* and *num_cols* are the number of SearchResult instances and number of columns, *arena_ids* the target arena ids

        There is one :class:`SearchResult` for each query fingerprint. The *arena_ids*
        are used to map the hit index back to the hit id.

        This constructor is not part of the public API.
        """
        if arena_ids is not None:
            if num_cols > len(arena_ids):
                raise ValueError("Not enough ids (%d) for the number of columns (%d)"
                                 % (len(arena_ids), num_cols))
        super(SearchResults, self).__init__(num_rows, num_cols, arena_ids)
        self._results = [SearchResult(self, i) for i in xrange(num_rows)]

    def __iter__(self):
        """Iterate over each SearchResult hit"""
        return iter(self._results)

    def __len__(self):
        """The number of rows in the SearchResults"""
        return super(SearchResults, self).__len__()

    @property
    def shape(self):
        """the tuple (number of rows, number of columns)

        The number of columns is the size of the target arena.
        """
        return (len(self), self._num_columns)

    def __getitem__(self, i):
        """Get the *i*-th SearchResult"""
        try:
            return self._results[i]
        except IndexError:
            raise IndexError("row index is out of range")

    def iter_indices(self):
        """For each hit, yield the list of target indices"""
        for i in xrange(len(self)):
            yield self._get_indices(i)

    def iter_ids(self):
        """For each hit, yield the list of target identifiers"""
        ids = self.target_ids
        for indicies in self.iter_indices():
            yield [ids[idx] for idx in indicies]

    def iter_scores(self):
        """For each hit, yield the list of target scores"""
        for i in xrange(len(self)):
            yield self._get_scores(i)

    def iter_indices_and_scores(self):
        """For each hit, yield the list of (target index, score) tuples"""
        for i in xrange(len(self)):
            yield zip(self._get_indices(i), self._get_scores(i))
    
    def iter_ids_and_scores(self):
        """For each hit, yield the list of (target id, score) tuples"""
        ids = self.target_ids
        for i in xrange(len(self)):
            yield [(ids[idx], score) for (idx, score) in self[i]]


    # I don't like how C-level doc strings can't report the call
    # signature even though keyword arguments are supported. I also
    # don't like maintaining the docstrings in C code.
    # Problem solved by interposing these Python methods
    def clear_all(self):
        """Remove all hits from all of the search results"""
        return super(SearchResults, self).clear_all()
            
    def count_all(self, min_score=None, max_score=None, interval="[]"):
        """Count the number of hits with a score between *min_score* and *max_score*
        
        Using the default parameters this returns the number of
        hits in the result.
        
        The default *min_score* of None is equivalent to -infinity.
        The default *max_score* of None is equivalent to +infinity.
        
        The *interval* parameter describes the interval end
        conditions. The default of "[]" uses a closed interval,
        where min_score <= score <= max_score. The interval "()"
        uses the open interval where min_score < score < max_score.
        The half-open/half-closed intervals "(]" and "[)" are
        also supported.
        
        :param min_score: the minimum score in the range.
        :type min_score: a float, or None for -infinity
        :param max_score: the maximum score in the range.
        :type max_score: a float, or None for +infinity
        :param interval: specify if the end points are open or closed.
        :type interval: one of "[]", "()", "(]", "[)"
        :returns: an integer count
        """
        return super(SearchResults, self).count_all(min_score, max_score, interval)
        
    def cumulative_score_all(self, min_score=None, max_score=None, interval="[]"):
        """The sum of all scores in all rows which are between *min_score* and *max_score*

        Using the default parameters this returns the sum of all of
        the scores in all of the results. With a specified range this
        returns the sum of all of the scores in that range. The
        cumulative score is also known as the raw score.
        
        The default *min_score* of None is equivalent to -infinity.
        The default *max_score* of None is equivalent to +infinity.
        
        The *interval* parameter describes the interval end
        conditions. The default of "[]" uses a closed interval,
        where min_score <= score <= max_score. The interval "()"
        uses the open interval where min_score < score < max_score.
        The half-open/half-closed intervals "(]" and "[)" are
        also supported.
        
        :param min_score: the minimum score in the range.
        :type min_score: a float, or None for -infinity
        :param max_score: the maximum score in the range.
        :type max_score: a float, or None for +infinity
        :param interval: specify if the end points are open or closed.
        :type interval: one of "[]", "()", "(]", "[)"
        :returns: a floating point count
        """
        return super(SearchResults, self).cumulative_score_all(min_score, max_score, interval)

    def reorder_all(self, order="decreasing-score"):
        """Reorder the hits for all of the rows based on the requested *order*.
        
        The available orderings are:

        * increasing-score - sort by increasing score
        * decreasing-score - sort by decreasing score
        * increasing-index - sort by increasing target index
        * decreasing-index - sort by decreasing target index
        * move-closest-first - move the hit with the highest score to the first position
        * reverse - reverse the current ordering
        
        :param ordering: the name of the ordering to use
        """
        return super(SearchResults, self).reorder_all(order)

    def to_csr(self, dtype=None):
        """Return the results as a SciPy compressed sparse row matrix.

        The returned matrix has the same shape as the SearchResult
        instance and can be passed into, for example, a scikit-learn
        clustering algorithm.
       
        By default the scores are stored with the `dtype` is "float64".

        This method requires that SciPy (and NumPy) be installed.
        
        :param dtype: a NumPy numeric data type
        :type dtype: string or NumPy type
        """
        import numpy as np
        import scipy.sparse

        if dtype is None:
            dtype = "float64"

        ## if reorder:
        ##     if reorder is True:
        ##         self.reorder_all("increasing-index")
        ##     else:
        ##         self.reorder_all(reorder)

        num_rows = len(self)
        max_columns = self._num_columns
        num_elements = sum(len(row) for row in self)
        
        # Figure out how many items will be in each of the arrays
        indptr = np.zeros(num_rows+1, "int32")
        indices = np.zeros(num_elements, "int32")
        data = np.zeros(num_elements, dtype)

        # Used to update the data fields
        indptr_value = 0
        indptr_offset = 1
        data_offset = 0
    
        # Fill in the data
        for row in self:
            column_indices = row.get_indices()
            column_scores = row.get_scores()
            num_columns = len(column_indices)

            # Update indptr
            indptr_value += len(column_indices)
            indptr[indptr_offset] = indptr_value
            indptr_offset += 1

            # update the indices and data
            indices[data_offset:data_offset+num_columns] = column_indices
            data[data_offset:data_offset+num_columns] = column_scores
            data_offset += num_columns

        return scipy.sparse.csr_matrix((data, indices, indptr), shape=(num_rows, max_columns),
                                       copy=False)
        
        
        
def _require_matching_fp_size(query_fp, target_arena):
    num_bytes = target_arena.metadata.num_bytes
    if num_bytes is None:
        num_bytes = target_arena.num_bytes
    if len(query_fp) != num_bytes:
        raise ValueError("query_fp uses %d bytes while target_arena uses %d bytes" % (
            len(query_fp), num_bytes))

def _require_matching_sizes(query_arena, target_arena):
    assert query_arena.metadata.num_bits is not None, "arenas must define num_bits"
    assert target_arena.metadata.num_bits is not None, "arenas must define num_bits"
    if query_arena.metadata.num_bits != target_arena.metadata.num_bits:
        raise ValueError("query_arena has %d bits while target_arena has %d bits" % (
            query_arena.metadata.num_bits, target_arena.metadata.num_bits))
    if query_arena.metadata.num_bytes != target_arena.metadata.num_bytes:
        raise ValueError("query_arena uses %d bytes while target_arena uses %d bytes" % (
            query_arena.metadata.num_bytes, target_arena.metadata.num_bytes))
    



def count_tanimoto_hits_fp(query_fp, target_arena, threshold=0.7):
    """Count the number of hits in *target_arena* at least *threshold* similar to the *query_fp*

    Example::
    
        query_id, query_fp = chemfp.load_fingerprints("queries.fps")[0]
        targets = chemfp.load_fingerprints("targets.fps")
        print chemfp.search.count_tanimoto_hits_fp(query_fp, targets, threshold=0.1)
        

    :param query_fp: the query fingerprint
    :type query_fp: a byte string
    :param target_arena: the target arena
    :type target_fp: a :class:`FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :returns: an integer count
    """
    _require_matching_fp_size(query_fp, target_arena)
    # Improve the alignment so the faster algorithms can be used
    query_start_padding, query_end_padding, query_fp = _chemfp.align_fingerprint(
        query_fp, target_arena.alignment, target_arena.storage_size)
                                                 
    counts = array.array("i", (0 for i in xrange(len(query_fp))))
    _chemfp.count_tanimoto_arena(threshold, target_arena.num_bits,
                                 query_start_padding, query_end_padding,
                                 target_arena.storage_size, query_fp, 0, 1,
                                 target_arena.start_padding, target_arena.end_padding,
                                 target_arena.storage_size, target_arena.arena,
                                 target_arena.start, target_arena.end,
                                 target_arena.popcount_indices,
                                 counts)
    return counts[0]


def count_tanimoto_hits_arena(query_arena, target_arena, threshold=0.7):
    """For each fingerprint in *query_arena*, count the number of hits in *target_arena* at least *threshold* similar to it
    
    Example::
    
        queries = chemfp.load_fingerprints("queries.fps")
        targets = chemfp.load_fingerprints("targets.fps")
        counts = chemfp.search.count_tanimoto_hits_arena(queries, targets, threshold=0.1)
        print counts[:10]

    The result is implementation specific. You'll always be able to
    get its length and do an index lookup to get an integer
    count. Currently it's a `ctypes array of longs <https://docs.python.org/2/library/ctypes.html#arrays>`_,
    but it could be an `array.array <https://docs.python.org/2/library/array.html>`_
    or Python list in the future.

    :param query_arena: The query fingerprints.
    :type query_arena: a :class:`chemfp.arena.FingerprintArena`
    :param target_arena: The target fingerprints.
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :returns: an array of counts
    """
    _require_matching_sizes(query_arena, target_arena)

    counts = (ctypes.c_int*len(query_arena))()
    _chemfp.count_tanimoto_arena(threshold, target_arena.num_bits,
                                 query_arena.start_padding, query_arena.end_padding,
                                 query_arena.storage_size,
                                 query_arena.arena, query_arena.start, query_arena.end,
                                 target_arena.start_padding, target_arena.end_padding,
                                 target_arena.storage_size,
                                 target_arena.arena, target_arena.start, target_arena.end,
                                 target_arena.popcount_indices,
                                 counts)
    return counts    

def count_tanimoto_hits_symmetric(arena, threshold=0.7, batch_size=100):
    """For each fingerprint in the *arena*, count the number of other fingerprints at least *threshold* similar to it

    A fingerprint never matches itself.

    The computation can take a long time. Python won't check check for
    a ``^C`` until the function finishes. This can be irritating. Instead,
    process only *batch_size* rows at a time before checking for a ``^C``.

    Note: the *batch_size* may disappear in future versions of chemfp.
    I can't detect any performance difference between the current value
    and a larger value, so it seems rather pointless to have. Let me
    know if it's useful to keep as a user-defined parameter.

    Example::

        arena = chemfp.load_fingerprints("targets.fps")
        counts = chemfp.search.count_tanimoto_hits_symmetric(arena, threshold=0.2)
        print counts[:10]

    The result object is implementation specific. You'll always be able to
    get its length and do an index lookup to get an integer
    count. Currently it's a ctype array of longs, but it could be an
    array.array or Python list in the future.

    :param arena: the set of fingerprints
    :type arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :param batch_size: the number of rows to process before checking for a ``^C``
    :type batch_size: integer
    :returns: an array of counts
    """
    N = len(arena)
    counts = (ctypes.c_int * N)()

    # This spends the entire time in C, which means ^C won't work until it finishes.
    # While it's theoretically slightly higher performance, I can't measure the
    # difference, and it's much better to let people be able to interrupt the program.
    #    _chemfp.count_tanimoto_hits_arena_symmetric(
    #        threshold, arena.num_bits,
    #        arena.start_padding, arena.end_padding, arena.storage_size, arena.arena,
    #        0, N, 0, N,
    #        arena.popcount_indices,
    #        counts)
    if batch_size <= 0:
        raise ValueError("batch_size must be positive")

    # Process N rows at a time, which lets Python handle ^C at times.
    # Since the code processes a triangle, this means that early
    # on there will be more time between ^C checks than later.
    # I'm not able to detect the Python overhead, so I'm not going
    # to make it more "efficient".
    for query_start in xrange(0, N, batch_size):
        query_end = min(query_start + batch_size, N)
        _chemfp.count_tanimoto_hits_arena_symmetric(
            threshold, arena.num_bits,
            arena.start_padding, arena.end_padding, arena.storage_size, arena.arena,
            query_start, query_end, 0, N,
            arena.popcount_indices,
            counts)

    return counts


def partial_count_tanimoto_hits_symmetric(counts, arena, threshold=0.7,
                                          query_start=0, query_end=None,
                                          target_start=0, target_end=None):
    """Compute a portion of the symmetric Tanimoto counts

    For most cases, use :func:`chemfp.search.count_tanimoto_hits_symmetric`
    instead of this function!
    
    This function is only useful for thread-pool implementations. In
    that case, set the number of OpenMP threads to 1.

    *counts* is a contiguous array of integers. It should be
    initialized to zeros, and reused for successive calls.

    The function adds counts for counts[*query_start*:*query_end*] based
    on computing the upper-triangle portion contained in the rectangle
    *query_start*:*query_end* and *target_start*:target_end* and using
    symmetry to fill in the lower half.

    You know, this is pretty complicated. Here's the bare minimum
    example of how to use it correctly to process 10 rows at a time
    using up to 4 threads::

        import chemfp
        import chemfp.search
        from chemfp import futures
        import array
        
        chemfp.set_num_threads(1)  # Globally disable OpenMP
        
        arena = chemfp.load_fingerprints("targets.fps")  # Load the fingerprints
        n = len(arena)
        counts = array.array("i", [0]*n)
        
        with futures.ThreadPoolExecutor(max_workers=4) as executor:
            for row in xrange(0, n, 10):
                executor.submit(chemfp.search.partial_count_tanimoto_hits_symmetric,
                                counts, arena, threshold=0.2,
                                query_start=row, query_end=min(row+10, n))
        
        print counts

    :param counts: the accumulated Tanimoto counts
    :type counts: a contiguous block of integer
    :param arena: the fingerprints.
    :type arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :param query_start: the query start row
    :type query_start: an integer
    :param query_end: the query end row
    :type query_end: an integer, or None to mean the last query row
    :param target_start: the target start row
    :type target_start: an integer
    :param target_end: the target end row
    :type target_end: an integer, or None to mean the last target row
    :returns: None
    """
    N = len(arena)
    
    if query_end is None:
        query_end = N
    elif query_end > N:
        query_end = N
        
    if target_end is None:
        target_end = N
    elif target_end > N:
        target_end = N

    if query_end > len(counts):
        raise ValueError("counts array is too small for the given query range")
    if target_end > len(counts):
        raise ValueError("counts array is too small for the given target range")

    _chemfp.count_tanimoto_hits_arena_symmetric(
        threshold, arena.num_bits,
        arena.start_padding, arena.end_padding, arena.storage_size, arena.arena,
        query_start, query_end, target_start, target_end,
        arena.popcount_indices,
        counts)


# These all return indices into the arena!

def threshold_tanimoto_search_fp(query_fp, target_arena, threshold=0.7):
    """Search for fingerprint hits in *target_arena* which are at least *threshold* similar to *query_fp*

    The hits in the returned :class:`chemfp.search.SearchResult` are in arbitrary order.

    Example::

        query_id, query_fp = chemfp.load_fingerprints("queries.fps")[0]
        targets = chemfp.load_fingerprints("targets.fps")
        print list(chemfp.search.threshold_tanimoto_search_fp(query_fp, targets, threshold=0.15))

    :param query_fp: the query fingerprint
    :type query_fp: a byte string
    :param target_arena: the target arena
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :returns: a :class:`chemfp.search.SearchResult`
    """
    _require_matching_fp_size(query_fp, target_arena)

    # Improve the alignment so the faster algorithms can be used
    query_start_padding, query_end_padding, query_fp = _chemfp.align_fingerprint(
        query_fp, target_arena.alignment, target_arena.storage_size)


    results = SearchResults(1, len(target_arena), target_arena.arena_ids)
    _chemfp.threshold_tanimoto_arena(
        threshold, target_arena.num_bits,
        query_start_padding, query_end_padding, target_arena.storage_size, query_fp, 0, 1,
        target_arena.start_padding, target_arena.end_padding,
        target_arena.storage_size, target_arena.arena,
        target_arena.start, target_arena.end,
        target_arena.popcount_indices,
        results, 0)
    return results[0]


def threshold_tanimoto_search_arena(query_arena, target_arena, threshold=0.7):
    """Search for the hits in the *target_arena* at least *threshold* similar to the fingerprints in *query_arena*

    The hits in the returned :class:`chemfp.search.SearchResults` are in arbitrary order.

    Example::
    
        queries = chemfp.load_fingerprints("queries.fps")
        targets = chemfp.load_fingerprints("targets.fps")
        results = chemfp.search.threshold_tanimoto_search_arena(queries, targets, threshold=0.5)
        for query_id, query_hits in zip(queries.ids, results):
            if len(query_hits) > 0:
                print query_id, "->", ", ".join(query_hits.get_ids())

    :param query_arena: The query fingerprints.
    :type query_arena: a :class:`chemfp.arena.FingerprintArena`
    :param target_arena: The target fingerprints.
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :returns: a :class:`chemfp.search.SearchResults`
    """
    _require_matching_sizes(query_arena, target_arena)

    num_queries = len(query_arena)
    num_targets = len(target_arena)

    results = SearchResults(num_queries, num_targets, target_arena.arena_ids)
    if num_queries:
        _chemfp.threshold_tanimoto_arena(
            threshold, target_arena.num_bits,
            query_arena.start_padding, query_arena.end_padding,
            query_arena.storage_size, query_arena.arena, query_arena.start, query_arena.end,
            target_arena.start_padding, target_arena.end_padding,
            target_arena.storage_size, target_arena.arena, target_arena.start, target_arena.end,
            target_arena.popcount_indices,
            results, 0)
    
    return results

def threshold_tanimoto_search_symmetric(arena, threshold=0.7, include_lower_triangle=True, batch_size=100):
    """Search for the hits in the *arena* at least *threshold* similar to the fingerprints in the arena

    When *include_lower_triangle* is True, compute the upper-triangle
    similarities, then copy the results to get the full set of
    results. When *include_lower_triangle* is False, only compute the
    upper triangle.

    The hits in the returned :class:`chemfp.search.SearchResults` are in arbitrary order.

    The computation can take a long time. Python won't check check for
    a ``^C`` until the function finishes. This can be irritating. Instead,
    process only *batch_size* rows at a time before checking for a ``^C``.

    Note: the *batch_size* may disappear in future versions of chemfp. Let
    me know if it really is useful for you to have as a user-defined parameter.

    Example::

        arena = chemfp.load_fingerprints("queries.fps")
        full_result = chemfp.search.threshold_tanimoto_search_symmetric(arena, threshold=0.2)
        upper_triangle = chemfp.search.threshold_tanimoto_search_symmetric(
                  arena, threshold=0.2, include_lower_triangle=False)
        assert sum(map(len, full_result)) == sum(map(len, upper_triangle))*2
                  
    :param arena: the set of fingerprints
    :type arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :param include_lower_triangle:
        if False, compute only the upper triangle, otherwise use symmetry to compute the full matrix
    :type include_lower_triangle: boolean
    :param batch_size: the number of rows to process before checking for a ^C
    :type batch_size: integer
    :returns: a :class:`chemfp.search.SearchResults`
    """
    
    if batch_size <= 0:
        raise ValueError("batch_size must be positive")
    N = len(arena)
    results = SearchResults(N, N, arena.arena_ids)

    if N:
        # Break it up into batch_size groups in order to let Python's
        # interrupt handler check for a ^C, which is otherwise
        # suppressed until the function finishes.
        for query_start in xrange(0, N, batch_size):
            query_end = min(query_start + batch_size, N)
            _chemfp.threshold_tanimoto_arena_symmetric(
                threshold, arena.num_bits,
                arena.start_padding, arena.end_padding, arena.storage_size, arena.arena,
                query_start, query_end, 0, N,
                arena.popcount_indices,
                results, query_start)

        if include_lower_triangle:
            _chemfp.fill_lower_triangle(results, N)
        
    return results



#def XXXpartial_threshold_tanimoto_search(results, query_arena, target_arena, threshold,
#                                      results_offsets=0):
#    pass

def partial_threshold_tanimoto_search_symmetric(results, arena, threshold=0.7,
                                                query_start=0, query_end=None,
                                                target_start=0, target_end=None,
                                                results_offset=0):
    """Compute a portion of the symmetric Tanimoto search results

    For most cases, use :func:`chemfp.search.threshold_tanimoto_search_symmetric`
    instead of this function!
    
    This function is only useful for thread-pool implementations. In
    that case, set the number of OpenMP threads to 1.

    *results* is a :class:`chemfp.search.SearchResults` instance which is at
    least as large as the arena. It should be reused for successive updates.

    The function adds hits to results[*query_start*:*query_end*], based
    on computing the upper-triangle portion contained in the rectangle
    *query_start*:*query_end* and *target_start*:*target_end*.

    It does not fill in the lower triangle. To get the full matrix,
    call *fill_lower_triangle*.

    You know, this is pretty complicated. Here's the bare minimum
    example of how to use it correctly to process 10 rows at a time
    using up to 4 threads::

        import chemfp
        import chemfp.search
        from chemfp import futures
        import array

        chemfp.set_num_threads(1)

        arena = chemfp.load_fingerprints("targets.fps")
        n = len(arena)
        results = chemfp.search.SearchResults(n, n, arena.ids)

        with futures.ThreadPoolExecutor(max_workers=4) as executor:
            for row in xrange(0, n, 10):
                executor.submit(chemfp.search.partial_threshold_tanimoto_search_symmetric,
                                results, arena, threshold=0.2,
                                query_start=row, query_end=min(row+10, n))

        chemfp.search.fill_lower_triangle(results)

    The hits in the :class:`chemfp.search.SearchResults` are in arbitrary order.

    :param results: the intermediate search results
    :type results: a :class:`chemfp.search.SearchResults` instance
    :param arena: the fingerprints.
    :type arena: a :class:`chemfp.arena.FingerprintArena`
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :param query_start: the query start row
    :type query_start: an integer
    :param query_end: the query end row
    :type query_end: an integer, or None to mean the last query row
    :param target_start: the target start row
    :type target_start: an integer
    :param target_end: the target end row
    :type target_end: an integer, or None to mean the last target row
    :param results_offset: use results[results_offset] as the base for the results
    :param results_offset: an integer
    :returns: None
    """
    assert arena.popcount_indices
    N = len(arena)
    
    if query_end is None:
        query_end = N
    elif query_end > N:
        query_end = N
        
    if target_end is None:
        target_end = N
    elif target_end > N:
        target_end = N

    if query_end > N:
        raise ValueError("counts array is too small for the given query range")
    if target_end > N:
        raise ValueError("counts array is too small for the given target range")

    if N:
        _chemfp.threshold_tanimoto_arena_symmetric(
            threshold, arena.num_bits,
            arena.start_padding, arena.end_padding, arena.storage_size, arena.arena,
            query_start, query_end, target_start, target_end,
            arena.popcount_indices,
            results, results_offset)


def fill_lower_triangle(results):
    """Duplicate each entry of *results* to its transpose

    This is used after the symmetric threshold search to turn the
    upper-triangle results into a full matrix.

    :param results: search results
    :type results: a :class:`chemfp.search.SearchResults`
    """
    _chemfp.fill_lower_triangle(results, len(results))



# These all return indices into the arena!

def knearest_tanimoto_search_fp(query_fp, target_arena, k=3, threshold=0.7):
    """Search for *k*-nearest hits in *target_arena* which are at least *threshold* similar to *query_fp*

    The hits in the :class:`chemfp.search.SearchResults` are ordered by
    decreasing similarity score.

    Example::

        query_id, query_fp = chemfp.load_fingerprints("queries.fps")[0]
        targets = chemfp.load_fingerprints("targets.fps")
        print list(chemfp.search.knearest_tanimoto_search_fp(query_fp, targets, k=3, threshold=0.0))

    :param query_fp: the query fingerprint
    :type query_fp: a byte string
    :param target_arena: the target arena
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :param k: the number of nearest neighbors to find.
    :type k: positive integer
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :returns: a :class:`chemfp.search.SearchResult`
    """
    _require_matching_fp_size(query_fp, target_arena)
    query_start_padding, query_end_padding, query_fp = _chemfp.align_fingerprint(
        query_fp, target_arena.alignment, target_arena.storage_size)
    
    if k < 0:
        raise ValueError("k must be non-negative")

    results = SearchResults(1, len(target_arena), target_arena.arena_ids)
    _chemfp.knearest_tanimoto_arena(
        k, threshold, target_arena.num_bits,
        query_start_padding, query_end_padding, target_arena.storage_size, query_fp, 0, 1,
        target_arena.start_padding, target_arena.end_padding,
        target_arena.storage_size, target_arena.arena, target_arena.start, target_arena.end,
        target_arena.popcount_indices,
        results, 0)
    _chemfp.knearest_results_finalize(results, 0, 1)

    return results[0]

def knearest_tanimoto_search_arena(query_arena, target_arena, k=3, threshold=0.7):
    """Search for the *k* nearest hits in the *target_arena* at least *threshold* similar to the fingerprints in *query_arena*

    The hits in the :class:`chemfp.search.SearchResults` are ordered by
    decreasing similarity score.

    Example::
    
        queries = chemfp.load_fingerprints("queries.fps")
        targets = chemfp.load_fingerprints("targets.fps")
        results = chemfp.search.knearest_tanimoto_search_arena(queries, targets, k=3, threshold=0.5)
        for query_id, query_hits in zip(queries.ids, results):
            if len(query_hits) >= 2:
                print query_id, "->", ", ".join(query_hits.get_ids())

    :param query_arena: The query fingerprints.
    :type query_arena: a :class:`chemfp.arena.FingerprintArena`
    :param target_arena: The target fingerprints.
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :param k: the number of nearest neighbors to find.
    :type k: positive integer
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :returns: a :class:`chemfp.search.SearchResults`
    """
    _require_matching_sizes(query_arena, target_arena)

    num_queries = len(query_arena)
    num_targets = len(target_arena)

    results = SearchResults(num_queries, num_targets, target_arena.arena_ids)

    _chemfp.knearest_tanimoto_arena(
        k, threshold, target_arena.num_bits,
        query_arena.start_padding, query_arena.end_padding,
        query_arena.storage_size, query_arena.arena, query_arena.start, query_arena.end,
        target_arena.start_padding, target_arena.end_padding,
        target_arena.storage_size, target_arena.arena, target_arena.start, target_arena.end,
        target_arena.popcount_indices,
        results, 0)
    
    _chemfp.knearest_results_finalize(results, 0, num_queries)
    
    return results


def knearest_tanimoto_search_symmetric(arena, k=3, threshold=0.7, batch_size=100):
    """Search for the *k*-nearest hits in the *arena* at least *threshold* similar to the fingerprints in the arena

    The hits in the :class:`SearchResults` are ordered by decreasing similarity score.

    The computation can take a long time. Python won't check check for
    a ``^C`` until the function finishes. This can be irritating. Instead,
    process only *batch_size* rows at a time before checking for a ``^C.``

    Note: the *batch_size* may disappear in future versions of chemfp. Let
    me know if it really is useful for you to keep as a user-defined parameter.

    Example::

        arena = chemfp.load_fingerprints("queries.fps")
        results = chemfp.search.knearest_tanimoto_search_symmetric(arena, k=3, threshold=0.8)
        for (query_id, hits) in zip(arena.ids, results):
            print query_id, "->", ", ".join(("%s %.2f" % hit) for hit in  hits.get_ids_and_scores())
    
    :param arena: the set of fingerprints
    :type arena: a :class:`chemfp.arena.FingerprintArena`
    :param k: the number of nearest neighbors to find.
    :type k: positive integer
    :param threshold: The minimum score threshold.
    :type threshold: float between 0.0 and 1.0, inclusive
    :param include_lower_triangle:
        if False, compute only the upper triangle, otherwise use symmetry to compute the full matrix
    :type include_lower_triangle: boolean
    :param batch_size: the number of rows to process before checking for a ^C
    :type batch_size: integer
    :returns: a :class:`chemfp.search.SearchResults`
    """
    N = len(arena)
    if batch_size <= 0:
        raise ValueError("batch_size must be positive")

    if not arena.popcount_indices:
        raise ValueError("knearest_tanimoto_search_symmetric requires a non-empty popcount_indices")

    results = SearchResults(N, N, arena.arena_ids)

    if N:
        # Break it up into batch_size groups in order to let Python's
        # interrupt handler check for a ^C, which is otherwise
        # suppressed until the function finishes.
        for query_start in xrange(0, N, batch_size):
            query_end = min(query_start + batch_size, N)
            _chemfp.knearest_tanimoto_arena_symmetric(
                k, threshold, arena.num_bits,
                arena.start_padding, arena.end_padding, arena.storage_size, arena.arena,
                arena.start+query_start, arena.start+query_end, arena.start, arena.end,
                arena.popcount_indices,
                results, query_start)
        _chemfp.knearest_results_finalize(results, 0, N)
    
    return results

# Start contains

def contains_fp(query_fp, target_arena):
    """Find the target fingerprints which contain the query fingerprint bits as a subset

    A target fingerprint contains a query fingerprint if all of the on
    bits of the query fingerprint are also on bits of the target
    fingerprint. This function returns a :class:`chemfp.search.SearchResult`
    containing all of the target fingerprints in *target_arena* that contain
    the *query_fp*.

    The SearchResult scores are all 0.0. 
    
    There is currently no direct way to limit the arena search range.
    Instead create a subarena by using Python's slice notation on the
    arena then search the subarena.

    :param query_fp: the query fingerprint
    :type query_fp: a byte string
    :param target_arena: The target fingerprints.
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :returns: a SearchResult instance
    """
    _require_matching_fp_size(query_fp, target_arena)
    results = SearchResults(1, len(target_arena), target_arena.arena_ids)

    # Improve the alignment so the faster algorithms can be used
    query_start_padding, query_end_padding, query_fp = _chemfp.align_fingerprint(
        query_fp, target_arena.alignment, target_arena.storage_size)

    _chemfp.contains_arena(
        target_arena.num_bits,
        query_start_padding, query_end_padding, target_arena.storage_size, query_fp, 0, 1,
        target_arena.start_padding, target_arena.end_padding,
        target_arena.storage_size, target_arena.arena,
        target_arena.start, target_arena.end,
        target_arena.popcount_indices,
        results, 0)
    return results[0]

    
def contains_arena(query_arena, target_arena):
    """Find the target fingerprints which contain the query fingerprints as a subset

    A target fingerprint contains a query fingerprint if all of the on
    bits of the query fingerprint are also on bits of the target
    fingerprint. This function returns a :class:`chemfp.search.SearchResults`
    where SearchResults[i] contains all of the target fingerprints in
    *target_arena* that contain the fingerprint for entry
    *query_arena* [i].
    
    The SearchResult scores are all 0.0.
    
    There is currently no direct way to limit the arena search range,
    though you can create and search a subarena by using Python's
    slice notation.

    :param query_arena: the query fingerprints
    :type query_arena: a :class:`chemfp.arena.FingerprintArena`
    :param target_arena: the target fingerprints
    :type target_arena: a :class:`chemfp.arena.FingerprintArena`
    :returns: a :class:`chemfp.search.SearchResults` instance, of the same size as query_arena
    """
    _require_matching_sizes(query_arena, target_arena)

    num_queries = len(query_arena)
    num_targets = len(target_arena)

    results = SearchResults(num_queries, num_targets, target_arena.arena_ids)
    if num_queries:
        _chemfp.contains_arena(
            target_arena.num_bits,
            query_arena.start_padding, query_arena.end_padding,
            query_arena.storage_size, query_arena.arena, query_arena.start, query_arena.end,
            target_arena.start_padding, target_arena.end_padding,
            target_arena.storage_size, target_arena.arena, target_arena.start, target_arena.end,
            target_arena.popcount_indices,
            results, 0)
    
    return results



"Create RDKit fingerprints"

# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB
# See the contents of "__init__.py" for full license details.

from __future__ import absolute_import, print_function

import os
import sys
import gzip

import rdkit
from rdkit import Chem
from rdkit.Chem import rdMolDescriptors
import rdkit.rdBase
from rdkit.Chem.MACCSkeys import GenMACCSKeys

try:
    from rdkit.Avalon import pyAvalonTools
    HAS_AVALON = True
except ImportError:
    pyAvalonTools = None
    HAS_AVALON = False

from . import sdf_reader
from .encodings import from_binary_lsb as _from_binary_lsb
from . import io
from . import types
from . import SOFTWARE as CHEMFP_SOFTWARE


# These are the things I consider to be public
__all__ = ["read_structures", "iter_smiles_molecules", "iter_sdf_molecules"]


# If the attribute doesn't exist then this is an unsupported pre-2010 RDKit distribution
SOFTWARE = "RDKit/" + getattr(rdkit.rdBase, "rdkitVersion", "unknown")

EXTERNAL_NAME = "RDKit"

# Used to check for version-dependent fingerprints
_VERSION_PROBE_MOL = Chem.MolFromSmiles(r"CC1=CC(=NN1CC(=O)NNC(=O)\C=C\C2=C(C=CC=C2Cl)F)C")

#########

# Helper function to convert a fingerprint to a sequence of bytes.

from rdkit import DataStructs
if getattr(DataStructs, "BitVectToBinaryText", None):
    _fp_to_bytes = DataStructs.BitVectToBinaryText
else:
    # Support for pre-2012 releases of RDKit
    def _fp_to_bytes(fp):
        return _from_binary_lsb(fp.ToBitString())[1]

#########
_allowed_formats = ["sdf", "smi"]
_format_extensions = {
    "sdf": "sdf",
    "mol": "sdf",
    "sd": "sdf",
    "mdl": "sdf",

    "smi": "smi",
    "can": "smi",
    "smiles": "smi",
    "ism": "smi",  # This does not exist in 2.0 or later
    "usm": "smi",  # Added in 2.0. Following OEChem's lead.
}


# While RDKit has a SMILES file parser, it doesn't handle reading from
# stdin or from compressed files. I wanted to support those as well, so
# ended up not using Chem.SmilesMolSupplier.

def iter_smiles_molecules(fileobj, close, errors="strict", location=None):
    """Iterate over the SMILES file records, returning (title, RDKit.Chem.Mol) pairs

    Each line of the input must at least one whitespace separated
    fields.  The first field is the SMILES. If there is a second field
    then it is used as the title, otherwise the title is the current
    record number, starting with "1".

    """
    name = getattr(fileobj, "name", None)
    error_handler = io.get_parse_error_handler(errors)
    if location is None:
        location = io.Location.from_source(name)

    
    it = _iter_smiles_molecules(fileobj, close, error_handler, location)
    next(it)
    return it

def _iter_smiles_molecules(fileobj, close, error_handler, location):
    recno = lineno = start_offset = end_offset = 0
    line = None
    mol = None
    def get_lineno():
        return lineno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return line
    def get_mol():
        return mol
    location.register(get_lineno = get_lineno,
                      get_recno = get_lineno,
                      get_offsets = get_offsets,
                      get_mol = get_mol,
        )
    yield "Ready!"
    try:
        for lineno, line in enumerate(fileobj, 1):
            start_offset = end_offset
            end_offset += len(line)
            
            words = line.split()
            if len(words) < 2:
                if not words:
                    mol = None
                    error_handler.error("Unexpected blank line", location)
                    continue
                id = None
            else:
                id = words[1]

            mol = Chem.MolFromSmiles(words[0])
            if mol is None:
                error_handler.error("Cannot parse the SMILES %r" % (words[0],), location)
                continue

            yield id, mol
    finally:
        if close is not None:
            close()
        location.save(lineno = get_lineno(),
                      recno = get_lineno(),
                      offsets = None,
                      record = None,
                      mol = None,
            )
    
        
def iter_sdf_molecules(fileobj, close, id_tag, errors="strict", location=None):
    """Iterate over the SD file records, returning (id, Chem.Mol) pairs

    """
    # If there's no explicit filename, see if fileobj has one
    name = getattr(fileobj, "name", None)
    error_handler = io.get_parse_error_handler(errors)
    if location is None:
        location = io.Location.from_source(name)

    return _iter_sdf_molecules(fileobj, close, id_tag, error_handler, location)

def _iter_sdf_molecules(fileobj, close, id_tag, error_handler, location):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    try:
        if id_tag is None:
            for i, text in enumerate(sdf_reader.iter_sdf_records(fileobj, close, error_handler, location)):
                mol = Chem.MolFromMolBlock(text)
                if mol is None:
                    # This was not a molecule?
                    error_handler.error("Could not parse molecule block", location)
                    continue
                title = mol.GetProp("_Name")
                id = title
                ## id = io.cleanup_ids(title)
                ## if not id:
                ##     error_handler.error("Missing title", location)
                ##     continue
                yield id, mol
        else:
            # According to
            #   http://www.mail-archive.com/rdkit-discuss@lists.sourceforge.net/msg01436.html
            # I can make a new SDMolSupplier, then SetData(), get the first record, and
            # get its property names. That's ... crazy.
            sdf_iter = sdf_reader.iter_sdf_records(fileobj, close, error_handler, location)
            for i, (id, text) in enumerate(sdf_reader.iter_tag_and_record(sdf_iter, id_tag)):
                mol = Chem.MolFromMolBlock(text)
                if mol is None:
                    # This was not a molecule?
                    error_handler.error("Could not parse molecule block", location)
                    continue
                ## if id is None:
                ##     error_handler.error("Missing id tag %r" % (id_tag,), location)
                ##     continue
                ## id = io.remove_special_characters_from_id(id)
                ## if not id:
                ##     error_handler.error("Empty id tag %r" % (id_tag,), location)
                ##     continue
                yield id, mol
    finally:
        location.save(mol=None)
        
            
# this class helps the case when someone is entering structure
# by-hand. (Most likely to occur with SMILES input). They would like
# to see the result as soon as a record is entered. But normal
# interation reader grabs a buffer of input to process, and not a
# line. It's faster that way. The following adapter supports the
# iterator protocol but turns it into simple readlines(). This will be
# slower but since do it only if stdin is a tty, there shouldn't be a
# problem.
## class _IterUsingReadline(object):
##     "Internal class for iterating a line at a time from tty input"
##     def __init__(self, fileobj):
##         self.fileobj = fileobj
##     def __iter__(self):
##         return iter(self.fileobj.readline, "")

## def _open(filename, compressed):
##     "Internal function to open the given filename, which might be compressed"
##     if filename is None:
##         if compressed:
##             return gzip.GzipFile(fileobj=sys.stdin, mode="r")
##         else:
##             # Python's iter reads a block.
##             # When someone types interactively, read only a line.
##             if sys.stdin.isatty():
##                 return _IterUsingReadline(sys.stdin)
##             else:
##                 return sys.stdin

##     if compressed:
##         return gzip.GzipFile(filename, "r")
##     return open(filename, "rU")

def is_valid_format(format):
    if format is None:
        return True
    try:
        format_name, compression = io.normalize_input_format(None, format, ("smi", None))
    except ValueError:
        return False
    format_name = _format_extensions.get(format_name, format_name)
    return format_name in ("sdf", "smi")
    

def read_structures(source, format=None, id_tag=None, reader_args=None, errors="strict", location=None):
    """Iterate the records in the input source as (title, RDKit.Chem.Mol) pairs

    'source' is a filename, a file object, or None for stdin
    'format' is either "sdf" or "smi" with optional ".gz" or ".bz2" extensions.
        If None then the format is inferred from the source extension
    'errors' is one of "strict" (default), "log", or "ignore" (other values are experimental)
    """
    format_name, compression = io.normalize_input_format(source, format, default=("smi", None))
    format_name = _format_extensions.get(format_name, format_name)
    if location is None:
        location = io.Location.from_source(source)
    location.save(record_format=format_name)
    
    error_handler = io.get_parse_error_handler(errors)
    if format_name == "sdf":
        # When chemfp-1.1 came out, it was faster for chemfp to parse the
        # records and have RDKit parse each record than to have RDKit handle
        # it. Here are the numbers:
        # """
        # I have an old PubChem file Compound_09425001_09450000.sdf .
        #   num. lines = 5,041,475   num. bytes = 159,404,037
        # 
        # Parse times for iter_sdf_records (parsing records in Python)
        #   37.6s (best of 37.6, 38.3, 37.8)
        # Parse times for the RDKit implementation (parsing records in C++)
        #   40.2s (best of 41.7, 41.33, 40.2)
        # 
        # The native RDKit reader is slower than the Python one and does
        # not have (that I can tell) support for compressed files, so
        # I'll go with the Python one. For those interested, here's the
        # RDKit version.
        # """
        
        # I re-ran a similar test for the chemfp-1.3 release. The
        # timing numbers are now about the same. The tradeoff now is
        # the improved error reporting in chemfp vs. the more accurate
        # parsing in RDKit. Since no one has complained, I'll stick
        # with using chemfp to identify the records.
        
        ## if (not compression) and (source is not None):
        ##     supplier = Chem.SDMolSupplier(source)
        ##     def native_sdf_reader():
        ##         for mol in supplier:
        ##             if mol is None:
        ##                 print >>sys.stderr, "Missing? after", title
        ##             else:
        ##                 title = mol.GetProp("_Name")
        ##                 yield title, mol
        ##     return native_sdf_reader()

        fileobj, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)
        # fileobj should always have the .name attribute set.
        return iter_sdf_molecules(fileobj, close, id_tag=id_tag,
                                  errors=error_handler, location=location)

    elif format_name == "smi":
        # I timed the native reader at 31.6 seconds (best of 31.6, 31.7, 31.7)
        # and the Python reader at 30.8 seconds (best of 30.8, 30.9, and 31.0)
        # Yes, the Python reader is faster and using it gives me better consistency
        #
        #if (not compressed) and (source is not None):
        #    supplier = Chem.SmilesMolSupplier(source, delimiter=" \t", titleLine=False)
        #    def native_smiles_reader():
        #        for mol in supplier:
        #            yield mol.GetProp("_Name"), mol
        #    return native_smiles_reader()
        fileobj, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)
        return iter_smiles_molecules(fileobj, close,
                                     errors=error_handler, location=location)

    else:
        raise ValueError("%s does not support the %r format"
                         % (EXTERNAL_NAME, format_name))
        

########### The topological fingerprinter

def _check_fromAtoms(fromAtoms):
    if fromAtoms is None:
        return
    if len(fromAtoms) == 0:
        raise ValueError("fromAtoms must contain at least one atom index")
    for atom_index in fromAtoms:
        if atom_index < 0:
            raise ValueError("fromAtoms must contain non-negative integers")
    return
    
# Some constants shared by the fingerprinter and the command-line code.

NUM_BITS = 2048
MIN_PATH = 1
MAX_PATH = 7
BITS_PER_HASH = 2
USE_HS = 1
assert USE_HS == 1, "Don't make this 0 unless you know what you are doing"

# Not supporting the tgtDensity and minSize options.
# This program generates fixed-length fingerprints.

def make_rdk_fingerprinter(minPath=MIN_PATH, maxPath=MAX_PATH, fpSize=NUM_BITS,
                           nBitsPerHash=BITS_PER_HASH, useHs=USE_HS,
                           fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be positive")
    if not (minPath > 0):
        raise ValueError("minPath must be positive")
    if not (maxPath >= minPath):
        raise ValueError("maxPath must not be smaller than minPath")
    if not (nBitsPerHash > 0):
        raise ValueError("nBitsPerHash must be positive")
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        def rdk_fingerprinter(mol):
            fp = Chem.RDKFingerprint(
                mol, minPath=minPath, maxPath=maxPath, fpSize=fpSize,
                nBitsPerHash=nBitsPerHash, useHs=useHs)
            return _fp_to_bytes(fp)
    else:
        def rdk_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = Chem.RDKFingerprint(
                mol, minPath=minPath, maxPath=maxPath, fpSize=fpSize,
                nBitsPerHash=nBitsPerHash, useHs=useHs,
                fromAtoms=new_fromAtoms)
            return _fp_to_bytes(fp)

            
    return rdk_fingerprinter

########### The MACCS fingerprinter


def maccs166_fingerprinter(mol):
    fp = GenMACCSKeys(mol)
    # In RDKit the first bit is always bit 1 .. bit 0 is empty (?!?!)
    bitstring_with_167_bits = fp.ToBitString()
    # I want the bits to start at 0, so I do a manual left shift
    return _from_binary_lsb(bitstring_with_167_bits[1:])[1]

def make_maccs166_fingerprinter():
    return maccs166_fingerprinter


########### The Morgan fingerprinter

# Some constants shared by the fingerprinter and the command-line code.

RADIUS = 2
USE_FEATURES = 0
USE_CHIRALITY = 0
USE_BOND_TYPES = 1

def make_morgan_fingerprinter(fpSize=NUM_BITS,
                              radius=RADIUS,
                              useFeatures=USE_FEATURES,
                              useChirality=USE_CHIRALITY,
                              useBondTypes=USE_BOND_TYPES,
                              fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be positive")
    if not (radius >= 0):
        raise ValueError("radius must be positive or zero")
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        def morgan_fingerprinter(mol):
            fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(
                mol, radius, nBits=fpSize, useChirality=useChirality,
                useBondTypes=useBondTypes, useFeatures=useFeatures)
            return _fp_to_bytes(fp)
    else:
        def morgan_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(
                mol, radius, nBits=fpSize, useChirality=useChirality,
                useBondTypes=useBondTypes, useFeatures=useFeatures,
                fromAtoms=new_fromAtoms)
            return _fp_to_bytes(fp)
        
    return morgan_fingerprinter


########### Torsion fingerprinter

TARGET_SIZE = 4

def make_torsion_fingerprinter(fpSize=NUM_BITS,
                               targetSize=TARGET_SIZE,
                               fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be positive")
    if not (targetSize >= 0):
        raise ValueError("targetSize must be positive or zero")
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        def torsion_fingerprinter(mol):
            fp = rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect(
                mol, nBits=fpSize, targetSize=targetSize)
            return _fp_to_bytes(fp)
    else:
        def torsion_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect(
                mol, nBits=fpSize, targetSize=targetSize,
                fromAtoms=new_fromAtoms)
            return _fp_to_bytes(fp)
    return torsion_fingerprinter

TORSION_VERSION = {
    "\xc2\x10@\x83\x010\x18\xa4,\x00\x80B\xc0\x00\x08\x00": "1",
    "\x13\x11\x103\x00\x007\x00\x00p\x01\x111\x0107": "2",
    }[make_torsion_fingerprinter(128)(_VERSION_PROBE_MOL)]

########### Atom Pair fingerprinter

MIN_LENGTH = 1
MAX_LENGTH = 30

def make_atom_pair_fingerprinter(fpSize=NUM_BITS,
                                 minLength=MIN_LENGTH,
                                 maxLength=MAX_LENGTH,
                                 fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be positive")
    if not (minLength >= 0):
        raise ValueError("minLength must be positive or zero")
    if not (maxLength >= minLength):
        raise ValueError("maxLength must not be less than minLength")
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        def pair_fingerprinter(mol):
            fp = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(
                mol, nBits=fpSize, minLength=minLength, maxLength=maxLength)
            return _fp_to_bytes(fp)
    else:
        def pair_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(
                mol, nBits=fpSize, minLength=minLength, maxLength=maxLength,
                fromAtoms=new_fromAtoms)
            return _fp_to_bytes(fp)
    return pair_fingerprinter

try:
    ATOM_PAIR_VERSION = {
        "\xfdB\xfe\xbd\xfa\xdd\xff\xf5\xff\x05\xdf?\xe3\xc3\xff\xfb": "1",
        "w\xf7\xff\xf7\xff\x17\x01\x7f\x7f\xff\xff\x7f\xff?\xff\xff": "2",
        }[make_atom_pair_fingerprinter(128)(_VERSION_PROBE_MOL)]
except Exception, err:
    # RDKit 2011.06 contained a bug
    if "Boost.Python.ArgumentError" in str(type(err)):
        ATOM_PAIR_VERSION = None
    else:
        raise


####################

from .types import FingerprintFamilyConfig, positive_int, nonnegative_int, zero_or_one

def _read_structures(metadata, source, format, id_tag, reader_args, errors, location):
    if metadata.aromaticity is not None:
        raise ValueError("RDKit does not support alternate aromaticity models "
                         "(want aromaticity=%r)" % metadata.aromaticity)
    return read_structures(source, format, id_tag, reader_args, errors, location)

# This configuration system is complicated because the design expected
# that all parameters will have the same defaults. Unfortunately,
# Avalon uses an fpSize default of 512 while the other fingerprints
# with an fpSize use 2048. Thus, I have a _base_fpsize used for the
# command-line args, which is able to figure out the correct default;
# a _base_avalon used for Avalon fingerprints; and _base_2048 used for
# everything else.

_base = FingerprintFamilyConfig(
    software = SOFTWARE + " " + CHEMFP_SOFTWARE,
    read_structures = _read_structures,
    )

# Only used for the command-line "--fpSize" option.
_base_fpsize = _base.clone()
_base_fpsize.add_argument(
    "fpSize", decoder=positive_int, metavar="INT", default=None, help=
        "number of bits in the fingerprint. Default of 2048 for RDK, Morgan, "
        "topological torsion, atom pair, and pattern fingerprints, and "
        "512 for Avalon fingerprints")

# Only used by Avalon
_base_avalon = _base.clone()
_base_avalon.add_argument(
    "fpSize", decoder=positive_int, metavar="INT", default=512, help=
        "number of bits in the fingerprint. Default of 2048 for RDK, Morgan, "
        "topological torsion, atom pair, and pattern fingerprints, and "
        "512 for Avalon fingerprints")

# Used by everything except Avalon
_base_2048 = _base.clone()
_base_2048.add_argument(
    "fpSize", decoder=positive_int, metavar="INT", default=2048, help=
        "number of bits in the fingerprint. Default of 2048 for RDK, Morgan, "
        "topological torsion, atom pair, and pattern fingerprints, and "
        "512 for Avalon fingerprints")


_base_2048.add_argument("minPath", decoder=positive_int, metavar="INT", default=MIN_PATH,
                   help = "minimum number of bonds to include in the subgraph")

_base_2048.add_argument("maxPath", decoder=positive_int, metavar="INT", default=MAX_PATH,
                   help = "maximum number of bonds to include in the subgraph")

_base_2048.add_argument("nBitsPerHash", decoder=positive_int, metavar="INT",
                   default=BITS_PER_HASH, help = "number of bits to set per path")

_base_2048.add_argument("useHs", decoder=zero_or_one, metavar="0|1", default=USE_HS,
                   help = "include information about the number of hydrogens on each atom")
# Morgan
_base_2048.add_argument("radius", decoder=nonnegative_int, metavar="INT", default=RADIUS,
                   help = "radius for the Morgan algorithm")

_base_2048.add_argument("useFeatures", decoder=zero_or_one, metavar="0|1",
                   default=USE_FEATURES, help = "use chemical-feature invariants")

_base_2048.add_argument("useChirality", decoder=zero_or_one, metavar="0|1",
                   default=USE_CHIRALITY, help = "include chirality information")

_base_2048.add_argument("useBondTypes", decoder=zero_or_one, metavar="0|1",
                   default=USE_BOND_TYPES, help = "include bond type information")


# torsion
_base_2048.add_argument("targetSize", decoder=positive_int, metavar="INT",
                   default=TARGET_SIZE, help = "number of bits in the fingerprint")

# pair
_base_2048.add_argument("minLength", decoder=nonnegative_int, metavar="INT",
                   default=MIN_LENGTH, help = "minimum bond count for a pair")

_base_2048.add_argument("maxLength", decoder=nonnegative_int, metavar="INT",
                   default=MAX_LENGTH, help = "maximum bond count for a pair")

# fromAtoms supported in rdkit, morgan, torsion, and atom_pair fingerprints

_base_2048.add_argument("fromAtoms", decoder=None, metavar="INT,INT,...",
                        default=None, help = "specify the atom indices to use")

# Avalon
IS_QUERY = 0
BIT_FLAGS = 15761407
_base_avalon.add_argument("isQuery", decoder=zero_or_one, metavar="0|1", default=IS_QUERY,
                   help="is the fingerprint for a query structure? (1 if yes, 0 if no)")

_base_avalon.add_argument("bitFlags", decoder=positive_int, metavar="INT", default=BIT_FLAGS,
                   help="bit flags, SSSBits are 32767 and similarity bits are 15761407")

#########

_MACCS_PROBE_MOL = Chem.MolFromSmiles("[He]")
_has_key_44 = GenMACCSKeys(_MACCS_PROBE_MOL)[44] != 0
MACCS_VERSION = {False: "1", True: "2"}[_has_key_44]

def _check_maccs_version(version):
    def make_fingerprinter(*args, **kwargs):
        if MACCS_VERSION != version:
            raise TypeError("This version of RDKit does not support the RDKit-MACCS/%s fingerprint" % (version,))
        return make_maccs166_fingerprinter(*args, **kwargs)
    return make_fingerprinter
            

RDKitMACCSFingerprintFamily_v1 = _base_2048.clone(
    name = "RDKit-MACCS166/1",
    num_bits = 166,
    make_fingerprinter = _check_maccs_version("1"),
    )
RDKitMACCSFingerprintFamily_v2 = _base_2048.clone(
    name = "RDKit-MACCS166/2",
    num_bits = 166,
    make_fingerprinter = _check_maccs_version("2"),
    )


# The number of bits depends on the parameters
def _get_num_bits(d):
    return d["fpSize"]

# Version 1.3.1 added support for fromAtoms. The type string needs to be
# backwards compatible to strings without 'fromAtoms=', so only add the
# term if fromAtoms is actually used.
class FormatFromAtoms(object):
    def __init__(self, format_str):
        self.format_str = format_str
    def __repr__(self):
        return "FormatFromAtoms(%r)" % (self.format_str,)
    def __mod__(self, d):
        # This is used to figure out what names are needed,
        # and to construct the format string.
        format_str = self.format_str
        fromAtoms = d["fromAtoms"]
        if fromAtoms is not None and not isinstance(fromAtoms, types.Dummy):
            format_str += " fromAtoms=%(fromAtoms)s"
            d["fromAtoms"] = ",".join(("%d" % i) for i in fromAtoms)
        return format_str % d

# Figure out the RDKit version
RDK_VERSION = {
    b"~\x89\x90\x04\xdb\xe9t\xc5\x87\xfe\xfc\xbf\xf8r\x01\xd6\xe9\xd7\xbbZv\xfeM\xc9\xfb\xf7{\xb7\xbd\xacl\xc9\xff\xb6s\xd5\x84\xf2\x8e\xdb\xcf\xcc\xb7\x02\xfa-\xa6\xee2\x1c\x19\x11\xe8q;\x08\x96NTN\xb2\xd7\x99\xa9\xca\xe7\xc3\x0c\x9bqP!\xd5\xe6\xf5\xff\x92\xf6\xed\xfa\xc4\xb4x\xf5\xd6\xde~\xfb\xcdp\xefv\xfb\xcd\xbb\x7f\xb9\xda\xf8?\xad\xd9\xf6\xe4\x94\xc3\xfb\xbf\xefv9\xff\xfe\x9bd\xf3\x15\x9d\x9bm\xeb\x0f\xaa$\xf5\x01\xef\xb7": "1",
    b"""\xcd\xd7\xb9\xc7\xd8\xa7\x18\n\xff\x90R\xc2\xfbS\xf5\xba\xd1\xf5\xf6F\x0f\x8d\xf1\x07\xbf\t_\x995`\xcaH\xf1>\x045q\'5\xce>\x8f\xf5\x18L\xb6\xae\xca\xde\xd5\xbd\xcd\xff/K"m\xdc\xec\xd3\x02\xb9\xf5\xc0\xb6_\xe2nv\xe36\xf5\xd5j\x1f\x90\xc6\xd1/\xb3]"=[\xbd\xb0F\xb9\xcfd\x96[e\xf7x\xafQ\xc7\xfa\xb3\xbf\xf3\xdfb\xb7]\xd7\xfe\xb7\xd6\xfd\x1f{\xd7\xbd\xf1\'P\xfaU\x89l^f\xde\xb5}\xaf""": "2",
}[make_rdk_fingerprinter(fpSize=1024)(_VERSION_PROBE_MOL)]


def _check_rdkit_fingerprint_version(version):
    def make_fingerprinter(*args, **kwargs):
        if RDK_VERSION != version:
            raise TypeError("This version of RDKit does not support the RDKit-Fingerprint/%s fingerprint" % (version,))
        return make_rdk_fingerprinter(*args, **kwargs)
    return make_fingerprinter

RDKitFingerprintFamily_v1 = _base_2048.clone(
    name = "RDKit-Fingerprint/1",
    format_string = ("minPath=%(minPath)s maxPath=%(maxPath)s fpSize=%(fpSize)s "
                     "nBitsPerHash=%(nBitsPerHash)s useHs=%(useHs)s"),
    num_bits = _get_num_bits,
    make_fingerprinter = _check_rdkit_fingerprint_version("1"),
    )

RDKitFingerprintFamily_v2 = _base_2048.clone(
    name = "RDKit-Fingerprint/2",
    format_string = FormatFromAtoms(
        "minPath=%(minPath)s maxPath=%(maxPath)s fpSize=%(fpSize)s "
        "nBitsPerHash=%(nBitsPerHash)s useHs=%(useHs)s"),
    num_bits = _get_num_bits,
    make_fingerprinter = _check_rdkit_fingerprint_version("2"),
    )

###

RDKitMorganFingerprintFamily_v1 = _base_2048.clone(
    name = "RDKit-Morgan/1",
    format_string = FormatFromAtoms(
             "radius=%(radius)d fpSize=%(fpSize)s useFeatures=%(useFeatures)d "
             "useChirality=%(useChirality)d useBondTypes=%(useBondTypes)d"),
    num_bits = _get_num_bits,
    make_fingerprinter = make_morgan_fingerprinter,
    )

###

def _check_torsion_version(version):
    def make_fingerprinter(*args, **kwargs):
        if TORSION_VERSION != version:
            raise TypeError("This version of RDKit does not support the RDKit-Torsion/%s fingerprint" % (version,))
        return make_torsion_fingerprinter(*args, **kwargs)
    return make_fingerprinter

RDKitTorsionFingerprintFamily_v1 = _base_2048.clone(
    name = "RDKit-Torsion/1",
    format_string = "fpSize=%(fpSize)s targetSize=%(targetSize)d",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_torsion_version("1"),
    )

RDKitTorsionFingerprintFamily_v2 = _base_2048.clone(
    name = "RDKit-Torsion/2",
    format_string = FormatFromAtoms("fpSize=%(fpSize)s targetSize=%(targetSize)d"),
    num_bits = _get_num_bits,
    make_fingerprinter = _check_torsion_version("2"),
    )

###

def _check_atom_pair_version(version):
    def make_fingerprinter(*args, **kwargs):
        if ATOM_PAIR_VERSION != version:
            raise TypeError("This version of RDKit does not support the RDKit-AtomPair/%s fingerprint" % (version,))
        return make_atom_pair_fingerprinter(*args, **kwargs)
    return make_fingerprinter

RDKitAtomPairFingerprintFamily_v1 = _base_2048.clone(
    name = "RDKit-AtomPair/1",
    format_string = "fpSize=%(fpSize)s minLength=%(minLength)d maxLength=%(maxLength)d",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_atom_pair_version("1"),
    )

RDKitAtomPairFingerprintFamily_v2 = _base_2048.clone(
    name = "RDKit-AtomPair/2",
    format_string = FormatFromAtoms("fpSize=%(fpSize)s minLength=%(minLength)d maxLength=%(maxLength)d"),
    num_bits = _get_num_bits,
    make_fingerprinter = _check_atom_pair_version("2"),
    )

########### Pattern fingerprinter

# PatternFingerprint((Mol) mol, (int) fpSize=2048, (list) atomCounts=[], (ExplicitBitVect) setOnlyBits=None)
# "NOTE: This function is experimental. The API or results may change from release to release."

def make_pattern_fingerprinter(fpSize=NUM_BITS):
    if not (fpSize > 0):
        raise ValueError("fpSize must be positive")
    def pattern_fingerprinter(mol):
        fp = Chem.PatternFingerprint(mol, fpSize=fpSize)
        return DataStructs.BitVectToBinaryText(fp)
    return pattern_fingerprinter

def _check_pattern_version():
    _pattern = make_pattern_fingerprinter(fpSize=256)(_VERSION_PROBE_MOL)
    if _pattern == b'\xfb\xce\xfe\xfd\xf8\xdb{l\x97\xfb\xfa\xf3\xfb\xff\xf4Y\xfc\x7f\xb9m\xff\xf7\xff\xff\xfe\x1fw\xef\xd3\xcfRy':
        return "1"

    _pattern2 = make_pattern_fingerprinter(fpSize=256)(Chem.MolFromSmiles("O"))
    
    if _pattern == b'{\xde\xfe\xfd\xfa\xdbsl\x97i\xfa\xb2\xfb\xff\xf49\xfd\x7f\xb9m\xff\xf3\xff\xff\xfe\x7fwo\xd3\x8f\xd2y':
        # The probe molecule isn't able to distinguish between v2 and v3
        if _pattern2 == b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return "2"
        if _pattern2 == b"\x00\x02\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return "3"

    if _pattern == b'{\xdf\xfe\xfd\xfa\xdfs|\x97i\xfa\xb2\xfb\xff\xf49\xfd\x7f\xb9o\xff\xf3\xff\xff\xfe\x7fwo\xdb\x8f\xd2y':
        # Might as well double-check _pattern2
        if _pattern2 == b"\x00\x02\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return "4"

    return "unknown"
    

if hasattr(Chem, "PatternFingerprint"):
    PATTERN_VERSION = _check_pattern_version()
else:
    PATTERN_VERSION = "unknown"


def _check_pattern_fingerprint(version):
    def make_fingerprinter(*args, **kwargs):
        if PATTERN_VERSION != version:
            raise TypeError("This version of RDKit does not support the RDKit-Pattern/%s fingerprint" % (version,))
        return make_pattern_fingerprinter(*args, **kwargs)
    return make_fingerprinter

RDKitPatternFingerprint_v1 = _base_2048.clone(
    name = "RDKit-Pattern/1",
    format_string = "fpSize=%(fpSize)s",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_pattern_fingerprint("1")
    )
    
RDKitPatternFingerprint_v2 = _base_2048.clone(
    name = "RDKit-Pattern/2",
    format_string = "fpSize=%(fpSize)s",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_pattern_fingerprint("2")
    )

# Added in RDKit version 2017.03.1.
RDKitPatternFingerprint_v3 = _base_2048.clone(
    name = "RDKit-Pattern/3",
    format_string = "fpSize=%(fpSize)s",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_pattern_fingerprint("3")
    )

# Added in RDKit version 2017.09.1.
RDKitPatternFingerprint_v4 = _base_2048.clone(
    name = "RDKit-Pattern/4",
    format_string = "fpSize=%(fpSize)s",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_pattern_fingerprint("4")
    )


########### Avalon fingerprinter

# GetAvalonFP( (object)mol [, (int)nBits=512 [, (bool)isQuery=False [, (bool)resetVect=False [, (int)bitFlags=15761407]]]])

# I don't know what the bitFlags do.
# avalonSSSBits        =    32767 = 0x007fff
# avalonSimilarityBits = 15761407 = 0xf07fff

AVALON_NBITS = 512

if HAS_AVALON:
    AVALON_VERSION = "1"
else:
    AVALON_VERSION = None

def make_avalon_fingerprinter(fpSize=AVALON_NBITS, isQuery=IS_QUERY, bitFlags=BIT_FLAGS):
    if not (fpSize > 0):
        raise ValueError("fpSize must be positive")
    isQuery = 1 if isQuery else 0
    def avalon_fingerprinter(mol):
        fp = pyAvalonTools.GetAvalonFP(mol, nBits=fpSize, isQuery=isQuery, bitFlags=bitFlags)
        return DataStructs.BitVectToBinaryText(fp)
    return avalon_fingerprinter

def _check_avalon_fingerprinter(version):
    def make_fingerprinter(*args, **kwargs):
        if not HAS_AVALON:
            raise TypeError("This version of RDKit is not compiled for Avalon support")
        if AVALON_VERSION != version:
            raise TypeError("This version of RDKit does not support the RDKit-Avalon/%s fingerprint" % (version,))
        return make_avalon_fingerprinter(*args, **kwargs)
    return make_fingerprinter

RDKitAvalonFingerprint_v1 = _base_avalon.clone(
    name = "RDKit-Avalon/1",
    format_string = "fpSize=%(fpSize)d isQuery=%(isQuery)s bitFlags=%(bitFlags)s",
    num_bits = _get_num_bits,
    make_fingerprinter = _check_avalon_fingerprinter("1"),
    )




# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import
import sys
import itertools
import textwrap

from .. import ChemFPError, open_fingerprint_writer
from .. import argparse, types, io
from .. import openeye as oe
from . import cmdsupport

##### Handle command-line argument parsing

# Build up some help text based on the atype and btype fields
atype_options = "\n  ".join(textwrap.wrap(" ".join(sorted(dict(oe._atype_flags)))))
btype_options = "\n  ".join(textwrap.wrap(" ".join(sorted(dict(oe._btype_flags)))))
from openeye.oegraphsim import (
    OEGetFPAtomType, OEFPAtomType_DefaultPathAtom,
    OEFPAtomType_DefaultCircularAtom, OEFPAtomType_DefaultTreeAtom,
    OEGetFPBondType, OEFPBondType_DefaultPathBond,
    OEFPBondType_DefaultCircularBond, OEFPBondType_DefaultTreeBond,
    )
type_help = """\
ATYPE is one or more of the following, separated by the '|' character
  %(atype_options)s
The following shorthand terms and expansions are also available:
 DefaultPathAtom = %(defaultpathatom)s
 DefaultCircularAtom = %(defaultcircularatom)s
 DefaultTreeAtom = %(defaulttreeatom)s
and 'Default' selects the correct value for the specified fingerprint.
Examples:
  --atype Default
  --atype Arom|AtmNum|FCharge|HCount

BTYPE is one or more of the following, separated by the '|' character
  %(btype_options)s
The following shorthand terms and expansions are also available:
 DefaultPathBond = %(defaultpathbond)s
 DefaultCircularBond = %(defaultcircularbond)s
 DefaultTreeBond = %(defaulttreebond)s
and 'Default' selects the correct value for the specified fingerprint.
Examples:
   --btype Default
   --btype Order|InRing

To simplify command-line use, a comma may be used instead of a '|' to
separate different fields. Example:
  --atype AtmNum,HvyDegree
""" % dict(atype_options=atype_options,
           btype_options=btype_options,
           defaultpathatom=OEGetFPAtomType(OEFPAtomType_DefaultPathAtom),
           defaultcircularatom=OEGetFPAtomType(OEFPAtomType_DefaultCircularAtom),
           defaulttreeatom=OEGetFPAtomType(OEFPAtomType_DefaultTreeAtom),
           defaultpathbond=OEGetFPBondType(OEFPBondType_DefaultPathBond),
           defaultcircularbond=OEGetFPBondType(OEFPBondType_DefaultCircularBond),
           defaulttreebond=OEGetFPBondType(OEFPBondType_DefaultTreeBond),

)


# Extra help text after the parameter descriptions
epilog = type_help + """\

OEChem guesses the input structure format based on the filename
extension and assumes SMILES for structures read from stdin.
Use "--in FORMAT" to select an alternative, where FORMAT is one of:
 
  File Type      Valid FORMATs (use gz if compressed)
  ---------      ------------------------------------
   SMILES        smi, ism, usm, can, smi.gz, ism.gz, can.gz
   SDF           sdf, mol, sdf.gz, mol.gz
   SKC           skc, skc.gz
   CDK           cdk, cdk.gz
   MOL2          mol2, mol2.gz
   PDB           pdb, ent, pdb.gz, ent.gz
   MacroModel    mmod, mmod.gz
   OEBinary v2   oeb, oeb.gz
"""

parser = argparse.ArgumentParser(
    description="Generate FPS fingerprints from a structure file using OEChem",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,    
    )

CircularFamily = oe.OpenEyeCircularFingerprintFamily_v2
path_group = parser.add_argument_group("path, circular, and tree fingerprints")
path_group.add_argument(
    "--path", action="store_true", help="generate path fingerprints (default)")
path_group.add_argument(
    "--circular", action="store_true", help="generate circular fingerprints")
path_group.add_argument(
    "--tree", action="store_true", help="generate tree fingerprints")

path_group.add_argument(
    "--numbits", action="store", type=int, metavar="INT", default=4096,
    help="number of bits in the fingerprint (default=4096)")
path_group.add_argument(
    "--minbonds", action="store", type=int, metavar="INT", default=0,
    help="minimum number of bonds in the path or tree fingerprint (default=0)")
path_group.add_argument(
    "--maxbonds", action="store", type=int, metavar="INT", default=None,
    help="maximum number of bonds in the path or tree fingerprint (path default=5, tree default=4)")
CircularFamily.add_argument_to_argparse("minradius", path_group)
CircularFamily.add_argument_to_argparse("maxradius", path_group)

# The expansion of 'Default' differs based on the fingerprint type
path_group.add_argument(
    "--atype", metavar="ATYPE", default="Default",
    help="atom type flags, described below (default=Default)")
path_group.add_argument(
    "--btype", metavar="BTYPE", default="Default",
    help="bond type flags, described below (default=Default)")

maccs_group = parser.add_argument_group("166 bit MACCS substructure keys")
maccs_group.add_argument(
    "--maccs166", action="store_true", help="generate MACCS fingerprints")

substruct_group = parser.add_argument_group("881 bit ChemFP substructure keys")
substruct_group.add_argument(
    "--substruct", action="store_true", help="generate ChemFP substructure fingerprints")

rdmaccs_group = parser.add_argument_group("ChemFP version of the 166 bit RDKit/MACCS keys")
rdmaccs_group.add_argument(
    "--rdmaccs", "--rdmaccs/2", action="store_true", help="generate 166 bit RDKit/MACCS fingerprints (version 2)")
rdmaccs_group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs")

parser.add_argument(
    "--aromaticity", metavar="NAME", choices=oe._aromaticity_flavor_names,
    default="openeye",
    help="use the named aromaticity model")

parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id (SD files only)")

parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default guesses from filename)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")

parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

cmdsupport.add_version(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")

def _get_atype_and_btype(args, atom_description_to_value, bond_description_to_value):
    try:
        atype = atom_description_to_value(args.atype)
    except ValueError, err:
        parser.error("--atype must contain '|' separated atom terms: %s" % (err,))
    try:
        btype = bond_description_to_value(args.btype)
    except ValueError, err:
        parser.error("--btype must contain '|' separated atom terms: %s" % (err,))
    return atype, btype

#######

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)

    supported_fingerprints = ("maccs166", "path", "substruct", "rdmaccs", "rdmaccs/1",
                              "circular", "tree")
    cmdsupport.mutual_exclusion(parser, args, "path", supported_fingerprints)

    if args.maccs166:
        # Create the MACCS keys fingerprinter
        opener = types.get_fingerprint_family("OpenEye-MACCS166")()
    elif args.path:
        if not (16 <= args.numbits <= 65536):
            parser.error("--numbits must be between 16 and 65536 bits")

        if not (0 <= args.minbonds):
            parser.error("--minbonds must be 0 or greater")
        if args.maxbonds is None:
            args.maxbonds = 5
        if not (args.minbonds <= args.maxbonds):
            parser.error("--maxbonds must not be smaller than --minbonds")
        atype, btype = _get_atype_and_btype(args, oe.path_atom_description_to_value,
                                            oe.path_bond_description_to_value)
        opener = types.get_fingerprint_family("OpenEye-Path")(
            numbits = args.numbits,
            minbonds = args.minbonds,
            maxbonds = args.maxbonds,
            atype = atype,
            btype = btype)
    elif args.circular:
        if not (16 <= args.numbits <= 65536):
            parser.error("--numbits must be between 16 and 65536 bits")

        if not (0 <= args.minradius):
            parser.error("--minradius must be 0 or greater")
        if not (args.minradius <= args.maxradius):
            parser.error("--maxradius must not be smaller than --minradius")
        atype, btype = _get_atype_and_btype(args, oe.circular_atom_description_to_value,
                                            oe.circular_bond_description_to_value)

        opener = types.get_fingerprint_family("OpenEye-Circular")(
            numbits = args.numbits,
            minradius = args.minradius,
            maxradius = args.maxradius,
            atype = atype,
            btype = btype)
    elif args.tree:
        if not (16 <= args.numbits <= 65536):
            parser.error("--numbits must be between 16 and 65536 bits")

        if not (0 <= args.minbonds):
            parser.error("--minbonds must be 0 or greater")
        if args.maxbonds is None:
            args.maxbonds = 4
        if not (args.minbonds <= args.maxbonds):
            parser.error("--maxbonds must not be smaller than --minbonds")
        atype, btype = _get_atype_and_btype(args, oe.tree_atom_description_to_value,
                                            oe.tree_bond_description_to_value)

        opener = types.get_fingerprint_family("OpenEye-Tree")(
            numbits = args.numbits,
            minbonds = args.minbonds,
            maxbonds = args.maxbonds,
            atype = atype,
            btype = btype)
        
    elif args.substruct:
        opener = types.get_fingerprint_family("ChemFP-Substruct-OpenEye")()
    elif args.rdmaccs:
        opener = types.get_fingerprint_family("RDMACCS-OpenEye")()
    elif getattr(args, "rdmaccs/1"):
        opener = types.get_fingerprint_family("RDMACCS-OpenEye/1")()
    else:
        parser.error("ERROR: fingerprint not specified?")

    if args.format is not None:
        if args.filenames:
            filename = args.filenames[0]
        else:
            filename = None
        if not oe.is_valid_format(filename, args.format):
            parser.error("Unsupported format specifier: %r" % (args.format,))

    if not oe.is_valid_aromaticity(args.aromaticity):
        parser.error("Unsupported aromaticity specifier: %r" % (args.aromaticity,))

    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")

    missing = cmdsupport.check_filenames(parser, None, args.format, args.filenames)

    # Ready the input reader/iterator
    metadata, reader = cmdsupport.read_multifile_structure_fingerprints(
        opener, args.filenames, format = args.format, id_tag = args.id_tag,
        reader_args = {"aromaticity": args.aromaticity}, errors = args.errors)
    
    location = io.Location.from_destination(args.output)
    try:
        writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
    except (IOError, ValueError) as err:
        msg = "Cannot open output fingerprint file: %s" % (err,)
        if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
            msg += ".\nUse --out to specify 'fps' or 'fps.gz'."
        parser.error(msg)
        
    try:
        with writer:
            writer.write_fingerprints(reader)
    except (ChemFPError, IOError, EOFError) as err:
        sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
        raise SystemExit(1)
    
if __name__ == "__main__":
    main()



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import, print_function

import sys
import re
import itertools

from .. import Metadata, FingerprintIterator, ChemFPError
from .. import argparse
from .. import encodings
from .. import sdf_reader
from .. import io
from .. import open_fingerprint_writer

from . import cmdsupport


def _check_num_bits(num_bits,  # from the user
                    fp_num_bits, # not None if the fp decoder know it exactly
                    num_bytes, # length of decoded fp in bytes
                    parser):
    """Check that the number of fingerprint bits and bytes match the user input

    Difficulties: some fingerprints have only a byte length, and the user
    doesn't have to specify the input.

    Returns the number of bits, or calls parser.error if there are problems
    """
    if fp_num_bits is not None:
        # The fingerprint knows exactly how many bits it contains
        if num_bits is None:
            # The user hasn't specified, so go with the exact number
            return fp_num_bits

        # If the user gave a value, make sure it matches
        if num_bits != fp_num_bits:
            parser.error(
                ("the first fingerprint has %(fp_num_bits)s bits which "
                 "is not the same as the --num-bits value of %(num_bits)s") % dict(
                    num_bits=num_bits, fp_num_bits=fp_num_bits))
            raise AssertionError("should not get here")
        
        return num_bits

    # If the number of bits isn't specified, assume it's exactly
    # enough to fill up the fingerprint bytes.
    if num_bits is None:
        return num_bytes * 8

    # The user specified the number of bits. The first fingerprint
    # has a number of bytes. This must be enough to hold the bits,
    # but only up to 7 bits larger.
    if (num_bits+7)//8 != num_bytes:
        parser.error(
            ("The byte length of the first fingerprint is %(num_bytes)s so --num-bits "
             "must be %(min)s <= num-bits <= %(max)s, not %(num_bits)s") % dict(
                num_bytes=num_bytes, min=num_bytes*8-7, max=num_bytes*8,
                num_bits=num_bits))
        raise AssertionError("should not get here")

    # Accept what the user suggested
    return num_bits

parser = argparse.ArgumentParser(
    description="Extract a fingerprint tag from an SD file and generate FPS fingerprints",
    #epilog=epilog,
    #formatter_class=argparse.RawDescriptionHelpFormatter,
    )

parser.add_argument("--id-tag", metavar="TAG", default=None,
            help="get the record id from TAG instead of the first line of the record")
parser.add_argument("--fp-tag", metavar="TAG", 
                    help="get the fingerprint from tag TAG (required)")

parser.add_argument("--in", metavar="FORMAT", dest="format", choices=["sdf", "sdf.gz"], default=None,
                    help="Specify if the input SD file is uncompressed or gzip compressed")
                    

parser.add_argument("--num-bits", metavar="INT", type=int,
                    help="use the first INT bits of the input. Use only when the "
                    "last 1-7 bits of the last byte are not part of the fingerprint. "
                    "Unexpected errors will occur if these bits are not all zero.")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="strict",
    help="how should structure parse errors be handled? (default=strict)")

parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")

parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")
parser.add_argument("--software", metavar="TEXT",
                    help="use TEXT as the software description")
parser.add_argument("--type", metavar="TEXT",
                    help="use TEXT as the fingerprint type description")

cmdsupport.add_version(parser)

parser.add_argument(
    "filenames", nargs="*", help="input SD files (default is stdin)", default=None)

# TODO:
# Currently I guess the format type based solely on the extension.
# Do I want to support encoding of the fps output?

#parser.add_argument(
#    "--compress", action="store", metavar="METHOD", default="auto",
#    help="use METHOD to compress the output (default='auto', 'none', 'gzip', 'bzip2')")


# This adds --cactvs, --base64 and other decoders to the command-line arguments
encodings._add_decoding_group(parser)

# Support the "--pubchem" option
shortcuts_group = parser.add_argument_group("shortcuts")

class AddSubsKeys(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.cactvs=True
        # the 1.3 is solely based on the version of the document at
        #  ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt
        namespace.software="CACTVS/unknown"
        namespace.type="CACTVS-E_SCREEN/1.0 extended=2"
        namespace.fp_tag="PUBCHEM_CACTVS_SUBSKEYS"

shortcuts_group.add_argument("--pubchem", nargs=0, action=AddSubsKeys,
   help = ("decode CACTVS substructure keys used in PubChem. Same as "
           "--software=CACTVS/unknown --type 'CACTVS-E_SCREEN/1.0 extended=2' "
           "--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --cactvs"))

###############


def read_sdf_ids_and_values(source, id_tag, fp_tag, compression, errors, location):
    sdf_iter = sdf_reader.open_sdf(source, compression, errors, location)
    assert fp_tag is not None, "how is fp_tag None?"
    if id_tag is None:
        # get title and fp_tag
        return sdf_reader.iter_title_and_tag(sdf_iter, fp_tag)
    return sdf_reader.iter_two_tags(sdf_iter, id_tag, fp_tag)

_illegal_value_pat = re.compile(r"[\000-\037]")

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)

    if not args.fp_tag:
        parser.error("argument --fp-tag is required")
    if args.num_bits is not None and args.num_bits <= 0:
        parser.error("--num-bits must be a positive integer")

    fp_decoder_name, fp_decoder = encodings._extract_decoder(parser, args)

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")
    
    cmdsupport.check_filenames(parser, None, "sdf", args.filenames)

    for attr in ("software", "type"):
        description = getattr(args, attr, None)
        if description is None:
            continue
        m = _illegal_value_pat.search(description)
        if m is None:
            continue
        parser.error("--%(attr)s description may not contain the character %(c)r" % dict(
                attr=attr, c = m.group(0)))

    error_handler = io.get_parse_error_handler(args.errors)

    # What follows is a bit tricky. I set up a chain of iterators:
    #   - iterate through the SDF iterators
    #   -   iterate through the (id, encoded_fp) pairs in each SDF iterator
    #   -     convert to (id, fp, num_bits) 3-element tuples
    #   -       use the first element to figure out the right metadata
    #   -       send to (id, fp) information to the io.write_fps1_output function


    # Iterate through each of the filenames, yielding the corresponding SDF iterator
    location = io.Location()

    if args.format == "sdf":
        compression = ""
    elif args.format == "sdf.gz":
        compression = "gz"
    else:
        compression = "auto"
    
    def get_ids_and_fps(id_tag, fp_tag):
        if not args.filenames:
            location.filename = "<stdin>"
            return read_sdf_ids_and_values(None, id_tag, fp_tag, compression=compression,
                                          errors=args.errors, location=location)
        
        if len(args.filenames) == 1:
            location.filename = args.filenames[0]
            return read_sdf_ids_and_values(args.filenames[0], id_tag, fp_tag,
                                           compression=compression,
                                           errors=args.errors, location=location)
        
        def multiple_filenames(filenames):    
            for filename in filenames:
                location.filename = filename
                for x in read_sdf_ids_and_values(filename, id_tag, fp_tag,
                                                 compression=compression,
                                                 errors=args.errors, location=location):
                    yield x
        return multiple_filenames(args.filenames)
    
    # Set up the error messages for missing id or fingerprints.
    if args.id_tag is None:
        MISSING_ID = "Missing title in the record starting %(where)s"
        MISSING_FP = "Missing fingerprint tag %(tag)r in record starting %(where)s"
    else:
        MISSING_ID = "Missing id tag %(tag)r in the record starting %(where)s"
        MISSING_FP = "Missing fingerprint tag %(tag)r in record %(id)r starting %(where)s"

    # This is either None or a user-specified integer
    num_bits = args.num_bits

    # At this point I don't have enough information to generate the metadata.
    # I won't get that until I've read the first record.
    outfile = None       # Don't open it until I'm ready to write the first record
    num_bytes = None     # Will need to get (or at least check) the fingerprint byte length

    # Decoded encoded fingerprints, yielding (id, fp, num_bits)
    
    def decode_fingerprints(encoded_fp_reader, error_handler):
        expected_num_bits = -1
        expected_fp_size = None
        
        for id, encoded_fp in encoded_fp_reader:
            assert id, "this check was supposed to be done in cmdsupport.cleanup_ids()"
            ## if not id:
            ##     msg = MISSING_ID % dict(id=id, where=location.where(),
            ##                             tag=args.id_tag)
            ##     error_handler.error(msg)
            ##     continue
            
            if not encoded_fp:
                msg = MISSING_FP % dict(id=id, where=location.where(),
                                        tag=args.fp_tag)
                error_handler.error(msg)
                continue

            # Decode the fingerprint, and complain if it isn't decodeable.
            try:
                num_bits, fp = fp_decoder(encoded_fp)
            except (ValueError, TypeError) as err:
                msg = ("Could not %(decoder_name)s decode tag %(tag)r value %(encoded_fp)s: %(err)s, %(where)s" %
                       dict(decoder_name=fp_decoder_name, tag=args.fp_tag,
                            where=location.where(), err=err, encoded_fp=repr(encoded_fp)))
                error_handler.error(msg)
                continue

            if num_bits != expected_num_bits:
                if expected_num_bits == -1:
                    expected_num_bits = num_bits
                else:
                    msg = ("Tag %(tag)r value %(encoded_fp)r has %(got)d bits but expected %(expected)d, %(where)s" %
                           dict(tag=args.fp_tag, encoded_fp=encoded_fp,
                                got=num_bits, expected=expected_num_bits,
                                where=location.where()))
                    error_handler.error(msg)
                    continue

            if len(fp) != expected_fp_size:
                if expected_fp_size is None:
                    expected_fp_size = len(fp)
                else:
                    msg = ("Tag %(tag)r value %(encoded_fp)r has %(got)d bytes but expected %(expected)d, %(where)s" %
                           dict(tag=args.fp_tag, encoded_fp=encoded_fp,
                                got=len(fp), expected=expected_fp_size,
                                where=location.where()))
                    error_handler.error(msg)
                    continue

            yield id, fp, num_bits



    ids_and_fps = get_ids_and_fps(args.id_tag, args.fp_tag)
    
    try:
        encoded_fps = cmdsupport.cleanup_ids(ids_and_fps, args.id_tag, error_handler, location)
        decoded_fps = decode_fingerprints(encoded_fps, error_handler)

        id, fp, num_bits = next(decoded_fps)
    except (ChemFPError, IOError) as err:
        sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
        raise SystemExit(1)
    except StopIteration:
        # No fingerprints? Make a new empty stream
        metadata = Metadata(date = io.utcnow())
        chained_reader = iter([])

    else:
        # Got the first fingerprint
        expected_num_bytes = len(fp)

        # Verify that they match
        expected_num_bits = _check_num_bits(args.num_bits, num_bits, expected_num_bytes, parser)
        

        chained_reader = itertools.chain( [(id, fp)], (x[:2] for x in decoded_fps) )
        metadata = Metadata(num_bits = expected_num_bits,
                            software = args.software,
                            type = args.type,
                            sources = args.filenames,
                            date = io.utcnow())

    output_location = io.Location.from_destination(args.output)
    try:
        writer = open_fingerprint_writer(args.output, metadata, format=args.output_format,
                                         location=output_location)
    except IOError as err:
        parser.error("Cannot open output file: %s" % (err,))
    except ValueError as err:
        msg = "Cannot open output fingerprint file: %s" % (err,)
        if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
            msg += ".\nUse --out to specify 'fps' or 'fps.gz'."
        parser.error(msg)

    try:
        with writer:
            writer.write_fingerprints(chained_reader)
    except ChemFPError as err:
        sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
        raise SystemExit(1)

if __name__ == "__main__":
    main()



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import with_statement, print_function

import re
import os
import sys
import binascii

from datetime import datetime

from . import ParseError

DEV_STDIN = "/dev/stdin"


def utcnow():
    return datetime.utcnow()
############ error handlers
def _get_text(msg, location):
    if location is None:
        return str(msg)
    return "%s, %s" % (msg, location.where())

class ErrorHandler(object):
    ## def warning(self, msg, location=None, extra=None):
    ##     sys.stderr.write("WARNING: %s\n" % _get_text(msg, location))

    def error(self, msg, location=None, extra=None):
        raise NotImplementedError("error() must be implemented in the subclass")
    

class IgnoreHandler(ErrorHandler):
    ## def warning(self, msg, location=None, extra=None):
    ##     pass
    
    def error(self, msg, location=None, extra=None):
        pass

class ReportHandler(ErrorHandler):
    ## def warning(self, msg, location=None, extra=None):
    ##     sys.stderr.write("WARNING: %s\n" % _get_text(msg, location))

    def error(self, msg, location=None, extra=None):
        sys.stderr.write("ERROR: %s. Skipping.\n" % _get_text(msg, location))

class LogHandler(ErrorHandler):
    def __init__(self, logger="chemfp"):
        if isinstance(logger, basestring):
            import logging
            logger = logging.getLogger(logger)
        self._logger = logger

    ## def warning(self, msg, location=None, extra=None):
    ##     self._logger.warning(_get_text(msg, location))

    def error(self, msg, location=None, extra=None):
        self._logger.error(_get_text(msg, location))

class StrictHandler(ErrorHandler):
    ## def warning(self, msg, location=None, extra=None):
    ##     pass

    def error(self, msg, location=None, extra=None):
        raise ParseError(msg, location), None, None


_error_handlers = {
    "ignore": IgnoreHandler(),
    "report": ReportHandler(),
    "strict": StrictHandler(),
}

def get_parse_error_handler(errors):
    if isinstance(errors, basestring):
        try:
            return _error_handlers[errors]
        except KeyError:
            pass
        if "log" not in _error_handlers:
            _error_handlers["log"] = handler = LogHandler("chemfp")
        if errors == "log":
            return handler
            
        raise ValueError("'errors' value must be an ErrorHandler or one of %s, not %r"
                         % (", ".join(map(repr, sorted(_error_handlers))), errors))

    return errors


############
    
_compression_extensions = {
    ".gz": "gz",
    ".gzip": "gz",
    ".bz2": "bz2",
    ".bzip": "bz2",
    ".bzip2": "bz2",
    ".xz": "xz",
    }

def _determine_format_from_filename(filename, default):
    # The filename could have 0, 1 or 2 extensions
    base, ext = os.path.splitext(filename)
    if ext == "":
        # No extensions, use the default
        return default
    ext = ext.lower()

    # If it's not a compression extension then it's a format indicator
    if ext not in _compression_extensions:
        # the [1:] is to remove the leading "."
        return (ext[1:], "")

    # Found a compression, now look for the format
    compression = _compression_extensions[ext]

    base, ext = os.path.splitext(base)
    if ext == "":
        # compression found but not the actual format type
        return (default[0], compression)

    # The [1:] is to remove the leading "."
    format_name = ext[1:].lower()
    
    return (format_name, compression)


    
def normalize_input_format(source, format, default=("fps", "")):
    if format is not None:
        # Either 0 or 1 dots
        terms = format.split(".")
        if terms:
            if len(terms) == 1:
                return terms[0], ""
            elif len(terms) == 2:
                return terms[0], terms[1]
        raise ValueError("Could not understand format specification %r" % (format,))
        
    if source is None:
        # Read from stdin
        filename = None
    elif isinstance(source, basestring):
        # This is a filename
        filename = source
    elif hasattr(source, "read"):
        # This is a Python file object
        filename = getattr(source, "name", None)
    else:
        raise ValueError("Unsupported source type %r" % (source,))

    if filename is None:
        # Reading from stdin or an unnamed file-like object with no
        # specified format. Not going to sniff the input. Instead, just
        return default

    return _determine_format_from_filename(filename, default)
        


def normalize_output_format(destination, format, default=("fps", "")):
    if format is not None:
        # Either 0 or 1 dots
        terms = format.split(".")
        if terms:
            if len(terms) == 1:
                return terms[0], ""
            elif len(terms) == 2:
                return terms[0], terms[1]
        raise ValueError("Could not understand format specification %r" % (format,))
    
    if destination is None:
        # Write to stdout
        filename = None
    elif isinstance(destination, (str, unicode)):
        # This is a filename
        filename = destination
    elif hasattr(destination, "write"):
        # This is a Python file object
        filename = getattr(destination, "name", None)
    else:
        raise ValueError("Unknown destination type %r" % (destination,))

    if filename is None:
        # Writing to stdout or an unnamed file-like object with no
        # specified format.
        return default
        
    return _determine_format_from_filename(filename, default)


def get_filename(source):
    if source is None:
        return None
    elif isinstance(source, basestring):
        return source
    else:
        return getattr(source, "name", None)

####

def _do_nothing():
    pass

def open_binary_output(destination, blame_name=None):
    if destination is None:
        return sys.stdout, _do_nothing
    if not isinstance(destination, basestring):
        return destination, _do_nothing
    base, ext = os.path.splitext(destination)
    ext = ext.lower()
    if ext not in _compression_extensions:
        f = open(destination, "wb")
        close = f.close
    else:
        f, close = open_compressed_output(destination, _compression_extensions[ext], blame_name)
    return f, close

def open_compressed_output(destination, compression, blame_name):
    # Raises a ValueError if the compression isn't supported
    if not compression:
        if destination is None:
            return sys.stdout, _do_nothing
        elif isinstance(destination, basestring):
            f = open(destination, "wb")
            return f, f.close
        else:
            return destination, _do_nothing

    if compression == "gz":
        import gzip
        if destination is None:
            f = gzip.GzipFile(mode="wb", fileobj=sys.stdout)
        elif isinstance(destination, basestring):
            f = gzip.open(destination, "wb")
        else:
            f = gzip.GzipFile(mode="wb", fileobj=destination)
        return f, f.close

    if compression == "bz2":
        try:
            import bz2
        except ImportError as err:
            raise ValueError("Please install the bz2 module for bzip2 compression support")
        
        if destination is None:
            if not os.path.exists("/dev/stdout"):
                raise ValueError("Python's bz2 library cannot write compressed data to stdout "
                               "on this platform")
            f = bz2.BZ2File("/dev/stdout", "wb")
        elif isinstance(destination, basestring):
            f = bz2.BZ2File(destination, "wb")
        else:
            raise ValueError("Python's bz2 library does not support writing to a file object")
        return f, f.close

    if compression == "xz":
        raise ValueError("chemfp does not yet support xz compression")

    if blame_name is None:
        raise ValueError("Unsupported compression type %r" % (compression,))
    else:
        raise ValueError("%s does not support compression type %r" % (blame_name, compression))

# The chemfp-1.1 release used 'open_compressed_input_universal' for all text input.
# This lets Python normalize "\n" or "\r\n" to "\n".
#
# However, the chemfp C extension for FPS parsing handles both newline
# conventions, and there's overhead for double-conversion. On my
# benchmark of 1M PubChem fingerprints.
#
# simsearch using --memory: 11% faster
#   Universal: 5.34, 5.22, 5.22  (chemfp-1.1)
#        byte: 4.72, 4.67, 4.79  (chemfp-1.3)
# simsearch using --scan: 40% faster
#   Universal: 1.35, 1.46, 1.45  (chemfp-1.1)
#        byte: 0.90, 0.81, 0.93  (chemfp-1.3)
# Using sum(1 for _ in chemfp.open("pubchem_million.fps")): 20% faster
#   Universal: 2.36, 2.35, 2.58  (chemfp-1.1)
#        byte: 1.87, 2.02, 2.02  (chemfp-1.3)
# Using chemfp.load_fingerprints("pubchem_million.fps"): 14% faster
#   Universal: 5.30, 5.18, 5.25  (chemfp-1.1)
#        byte: 4.43, 4.65, 4.67  (chemfp-1.3)

def open_compressed_input(source, compression, blame_name):
    if not compression:
        if source is None:
            return sys.stdin, _do_nothing
        elif isinstance(source, basestring):
            f = open(source, "rb")
            return f, f.close
        else:
            return source, _do_nothing

    if compression == "gz":
        import gzip
        if source is None:
            f = gzip.GzipFile(fileobj=sys.stdin)
        elif isinstance(source, basestring):
            f = gzip.open(source, "rb")
        else:
            f = gzip.GzipFile(mode="rb", fileobj=source)
        return f, f.close

    if compression == "bz2":
        try:
            import bz2
        except ImportError as err:
            raise ValueError("Please install the bz2 module for bzip2 compression support")
        if source is None:
            # bz2 doesn't support Python objects. On some platforms
            # I can work around the problem
            if not os.path.exists("/dev/stdin"):
                raise NotImplementedError("Cannot compressed bzip2 data from stdin on this platform")
            f = bz2.BZ2File("/dev/stdin", "rb")
        elif isinstance(source, basestring):
            f = bz2.BZ2File(source, "rb")
        else:
            # Well, I could emulate it, but I'm not going to
            raise NotImplementedError("bzip decompression from file-like objects is not supported")
        return f, f.close

    if compression == "xz":
        raise NotImplementedError("xz decompression is not supported")

    if blame_name is None:
        raise ValueError("Unsupported compression type %r" % (compression,))
    else:
        raise ValueError("%s does not support compression type %r" % (blame_name, compression))


def write_fps1_magic(outfile):
    outfile.write("#FPS1\n")

def write_fps1_header(outfile, metadata):
    if metadata is None:
        return 0
    lines = []
    if metadata.num_bits is not None:
        lines.append("#num_bits=%d\n" % metadata.num_bits)

    if metadata.type is not None:
        assert "\n" not in metadata.type
        lines.append("#type=" + metadata.type.encode("ascii")+"\n") # type cannot contain non-ASCII characters

    if metadata.software is not None:
        assert "\n" not in metadata.software
        lines.append("#software=" + metadata.software.encode("utf8")+"\n")

    if metadata.aromaticity is not None:
        lines.append("#aromaticity=" + metadata.aromaticity.encode("ascii") + "\n")

    for source in metadata.sources:
        # Ignore newlines in the source filename, if present
        source = source.replace("\n", "")
        lines.append("#source=" + source.encode("utf8")+"\n")

    if metadata.date is not None:
        date = metadata.date
        if not isinstance(date, basestring):
            date = date.replace(microsecond=0).isoformat()
        lines.append("#date=" + date.encode("ascii")+"\n") # date cannot contain non-ASCII characters

    outfile.writelines(lines)
    return len(lines)

def write_fps1_fingerprint(outfile, fp, id):
    if "\t" in id:
        raise ValueError("Fingerprint ids must not contain a tab: %r" % (id,))
    if "\n" in id:
        raise ValueError("Fingerprint ids must not contain a newline: %r" % (id,))
    if not id:
        raise ValueError("Fingerprint ids must not be the empty string")
    
    outfile.write("%s\t%s\n" % (binascii.hexlify(fp), id))


# This is a bit of a hack. If I open a file then I want to close it,
# but if I use stdout then I don't want to close it.

## class _closing_output(object):
##     def __init__(self, destination):
##         self.destination = destination
##         self.output = open_output(destination)
##     def __enter__(self):
##         return self.output
##     def __exit__(self, *exec_info):
##         if isinstance(self.destination, basestring):
##             self.output.close()

## def write_fps1_output(reader, destination, metadata=None):
##     if metadata is None:
##         metadata = reader.metadata
##     hexlify = binascii.hexlify
##     with _closing_output(destination) as outfile:
##         with ignore_pipe_errors:
##             write_fps1_magic(outfile)
##             write_fps1_header(outfile, metadata)

##             for i, (id, fp) in enumerate(reader):
##                 if "\t" in id:
##                     raise ValueError("Fingerprint ids must not contain a tab: %r in record %d" %
##                                      (id, i+1))
##                 if "\n" in id:
##                     raise ValueError("Fingerprint ids must not contain a newline: %r in record %d" %
##                                      (id, i+1))
##                 if not id:
##                     raise ValueError("Fingerprint ids must not be the empty string in record %d" %
##                                      (i+1,))
##                 outfile.write("%s\t%s\n" % (hexlify(fp), id))


_where_template = {
    (0, 0, 0): "<unknown position>",
    (0, 0, 1): "record #%(recno)d",
    (0, 1, 0): "line %(lineno)d",
    (0, 1, 1): "line %(lineno)d, record #%(recno)d",
    (1, 0, 0): "file %(filename)r",
    (1, 0, 1): "file %(filename)r, record #%(recno)d",
    (1, 1, 0): "file %(filename)r, line %(lineno)d",
    (1, 1, 1): "file %(filename)r, line %(lineno)d, record #%(recno)d",
    }

class Location(object):
    """Get location and other internal reader and writer state information

    A Location instance gives a way to access information like
    the current record number, line number, and molecule object.::

      >>> import chemfp
      >>> with chemfp.read_molecule_fingerprints("RDKit-MACCS166",
      ...                        "ChEBI_lite.sdf.gz", id_tag="ChEBI ID") as reader:
      ...   for id, fp in reader:
      ...     if id == "CHEBI:3499":
      ...         print("Record starts at line", reader.location.lineno)
      ...         print("Record byte range:", reader.location.offsets)
      ...         print("Number of atoms:", reader.location.mol.GetNumAtoms())
      ...         break
      ... 
      [08:18:12]  S group MUL ignored on line 103
      Record starts at line 3599
      Record byte range: (138171, 141791)
      Number of atoms: 36

    The supported properties are:

      * filename - a string describing the source or destination
      * lineno - the line number for the start of the file
      * mol - the toolkit molecule for the current record
      * offsets - the (start, end) byte positions for the current record
      * output_recno - the number of records written successfully
      * recno - the current record number
      * record - the record as a text string
      * record_format - the record format, like "sdf" or "can"
       

    Most of the readers and writers do not support all of the properties.
    Unsupported properties return a None. The *filename* is a read/write
    attribute and the other attributes are read-only.
    
    If you don't pass a location to the readers and writers then they will
    create a new one based on the source or destination, respectively.
    You can also pass in your own Location, created as ``Location(filename)``
    if you have an actual filename, or ``Location.from_source(source)`` or
    ``Location.from_destination(destination)`` if you have a more generic
    source or destination.

    """
    _get_recno = None
    _get_output_recno = None
    _get_lineno = None
    _get_offsets = None
    _get_mol = None
    _get_record = None
    _record_format = None
    
    def __init__(self, filename=None):
        """Use *filename* as the location's filename"""
        self.filename = filename

    def __repr__(self):
        """Return a string like 'Location("<stdout>")'"""
        return "Location(%r)" % (self.filename,)

    def where(self):
        """Return a human readable description about the current reader or writer state.

        The description will contain the filename, line number, record
        number, and up to the first 40 characters of the first line of
        the record, if those properties are available.
        """
        filename = self.filename
        lineno = self.lineno
        recno = self.recno

        template = _where_template[ (filename is not None, lineno is not None, recno is not None) ]
        s = template % {"filename": filename, "lineno": lineno, "recno": recno}

        first_line = self.first_line

        if first_line:  # Don't show None and don't show blank lines
            if len(first_line) > 40:
                t = repr(first_line[:40])
                t = t[:-1] + " ..." + t[-1]
                s += ": first line starts %s" % (t,)
            else:
                s += ": first line is %s" % (repr(first_line),)
        return s

    @classmethod
    def from_source(cls, source):
        """Create a Location instance based on the source

        If *source* is a string then it's used as the filename.
        If *source* is None then the location filename is "<stdin>".
        If *source* is a file object then its ``name`` attribute
        is used as the filename, or None if there is no attribute.
        """
        if source is None:
            return cls("<stdin>")
        if isinstance(source, basestring):
            return cls(source)
        return cls(getattr(source, "name", None))

    @classmethod
    def from_destination(cls, destination):
        """Create a Location instance based on the destination
        
        If *destination* is a string then it's used as the filename.
        If *destination* is None then the location filename is "<stdout>".
        If *destination* is a file object then its ``name`` attribute
        is used as the filename, or None if there is no attribute.
        """
        if destination is None:
            return cls("<stdout>")
        if isinstance(destination, basestring):
            return cls(destination)
        return cls(getattr(destination, "name", None))


    def clear_registry(self):
        """Part of the internal API, and subject to change."""
        self._get_recno = None
        self._get_output_recno = None
        self._get_lineno = None
        self._get_offsets = None
        self._get_mol = None
        self._get_record = None

    def register(self, **kwargs):
        """Part of the internal API, and subject to change."""
        for k, v in kwargs.items():
            if k in ("get_recno", "get_output_recno", "get_lineno", "get_offsets", "get_mol", "get_record"):
                setattr(self, "_" + k, v)
            else:
                raise KeyError(k)

    def save(self, **kwargs):
        """Part of the internal API, and subject to change."""
        for k, v in kwargs.items():
            if k in ("recno", "output_recno", "lineno", "offsets", "mol", "record"):
                def recall_value(value=v):
                    return value
                setattr(self, "_get_" + k, recall_value)
            elif k == "record_format":
                self._record_format = v
            else:
                raise KeyError(k)

    def get_registry(self):
        """Part of the internal API, and subject to change."""
        return {"get_recno": self._get_recno,
                "get_output_recno": self._get_output_recno,
                "get_lineno": self._get_lineno,
                "get_offsets": self._get_offsets,
                "get_mol": self._get_mol,
                "get_record": self._get_record,
                }

    @property
    def recno(self):
        """The current record number

        For writers this is the number of records sent to
        the writer, and output_recno is the number of records
        sucessfully written to the file or string.
        """
        _get_recno = self._get_recno
        if _get_recno is None:
            return None
        return _get_recno()

    @property
    def output_recno(self):
        """The number of records actually written to the file or string.

        The value ``recno - output_recno`` is the number of records
        sent to the writer but which had an error and could not be
        written to the output.
        """
        _get_output_recno = self._get_output_recno
        if _get_output_recno is None:
            return None
        return _get_output_recno()

    @property
    def lineno(self):
        """The current line number, starting from 1"""
        _get_lineno = self._get_lineno
        if _get_lineno is None:
            return None
        return _get_lineno()

    @property
    def offsets(self):
        """The (start, end) byte offsets, starting from 0

        *start* is the record start byte position and *end* is
        one byte past the last byte of the record.
        """
        _get_offsets = self._get_offsets
        if _get_offsets is None:
            return None
        return _get_offsets()

    @property
    def mol(self):
        """The molecule object for the current record"""
        _get_mol = self._get_mol
        if _get_mol is None:
            return None
        return _get_mol()

    @property
    def record(self):
        """The current record as an uncompressed text string"""
        _get_record = self._get_record
        if _get_record is None:
            return None
        return _get_record()

    @property
    def record_format(self):
        """The record format name"""
        return self._record_format

    @property
    def first_line(self):
        """The first line of the current record"""
        _get_record = self._get_record
        if _get_record is None:
            return None
        record = _get_record()
        if record is None:
            return None
        first_line, _, _ = record.partition(b"\n")
        return first_line
        #return first_line.rstrip(b"\r").decode("utf8", errors="replace")

    @property
    def first_line_bytes(self):
        """The first line of the current record"""
        _get_record = self._get_record
        if _get_record is None:
            return None
        record = _get_record()
        if record is None:
            return None
        first_line, _, _ = record.partition(b"\n")
        return first_line.rstrip(b"\r")
        

    def get_info(self):
        """Part of the internal API, and subject to change."""
        d = {}
        for k, v in self.get_registry().items():
            if v is not None:
                d[k] = v
        return d




# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import, print_function

import datetime
from cStringIO import StringIO
from __builtin__ import open as _builtin_open
import binascii
import _chemfp
import re
import sys
import heapq
import itertools
from binascii import hexlify as _hexlify

from . import Metadata, ParseError, FingerprintReader, FingerprintWriter
from . import fps_search
from . import io

# I tried a wide range of sizes for my laptop, with both compressed
# and uncompressed files, and found that the best size was around
# 2**17. Actually, 2**16.8 was the absolute best, which gives
BLOCKSIZE=11400
# (BTW, the compressed time took 1.3x the uncompressed time)

def open_fps(source, format_name, compression, location=None):
    if format_name != "fps":
        raise ValueError("Unknown format %r" % (format_name,))

    infile, close = io.open_compressed_input(source, compression, "chemfp")
    if location is None:
        location = io.Location.from_source(source)
    #filename = io.get_filename(source)
    metadata, block = read_header(infile, location)
    return FPSReader(infile, close, metadata, location, block)


# This never buffers
def _read_blocks(infile):
    while 1:
        block = infile.read(BLOCKSIZE)
        if not block:
            break
        if block[-1:] == "\n":
            yield block
            continue
        line = infile.readline()
        if not line:
            # Note: this might not end with a newline!
            yield block
            break
        yield block + line

            

class FPSReader(FingerprintReader):
    """FPS file reader

    This class implements the :class:`chemfp.FingerprintReader` API. It
    is also its own a context manager, which automatically closes the
    file when the manager exists.

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.Metadata` instance with information about the fingerprint type
       
    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance with parser location and state information
       
    .. py:attribute:: closed

       True if the file is open, else False
    
    The FPSReader.location only tracks the "lineno" variable.
    """
    _search = fps_search
    def __init__(self, infile, close, metadata, location, first_fp_block):
        self._infile = infile
        self._close = close
        self._filename = getattr(infile, "name", "<unknown>")
        self.metadata = metadata
        self.location = location
        self._first_fp_block = first_fp_block
        if metadata.num_bytes is None:
            self._expected_hex_len = 0
        else:
            self._expected_hex_len = 2*metadata.num_bytes
        self._hex_len_source = "size in header"

        self._at_start = True
        self._it = None
        self._block_reader = None
        self._iter = None

        self._record_iter = None

        self.closed = False

    def close(self):
        """Close the file"""
        if self._close is not None:
            self._close()
        self.closed = True

    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
    
    def iter_blocks(self):
        """This is not part of the public API"""
        if self._block_reader is None:
            self._block_reader = iter(self._iter_blocks())
        return self._block_reader

    def _iter_blocks(self):
        if not self._at_start:
            raise TypeError("Already iterating")
        
        self._at_start = False

        if self._first_fp_block is None:
            return
        
        block_stream = _read_blocks(self._infile)
        yield self._first_fp_block
        for block in block_stream:
            yield block

    def iter_rows(self):
        """This is not part of the public API"""
        unhexlify = binascii.unhexlify
        expected_hex_len = self._expected_hex_len
        
        location = self.location
        lineno = location.lineno
        def get_lineno():
            return lineno
        location.register(get_lineno=get_lineno)
            
        try:
            for block in self.iter_blocks():
                for line in block.splitlines(True):
                    lineno += 1
                    err = _chemfp.fps_line_validate(expected_hex_len, line)
                    if err:
                        raise ParseError(_chemfp.strerror(err), location)
                    yield line[:-1].split(b"\t")
        finally:
            location.save(lineno=lineno)

    def __next__(self):
        if self._record_iter is None:
            self._record_iter = self._make_record_iter()
        return next(self._record_iter)

    def next(self):
        "Return the next (id, fp) pair"
        if self._record_iter is None:
            self._record_iter = self._make_record_iter()
        return next(self._record_iter)
        

    def __iter__(self):
        "Iterate through the (id, fp) pairs"
        if self._record_iter is None:
            self._record_iter = iter(self._make_record_iter())
        return self._record_iter

    def _make_record_iter(self):
        unhexlify = binascii.unhexlify
        expected_hex_len = self._expected_hex_len
        
        location = self.location
        lineno = location.lineno
        def get_lineno():
            return lineno
        location.register(get_lineno=get_lineno)

        try:
            for block in self.iter_blocks():
                for line in block.splitlines(True):
                    lineno += 1
                    err, id_fp = _chemfp.fps_parse_id_fp(expected_hex_len, line)
                    if err:
                        # Include the line?
                        raise ParseError(_chemfp.strerror(err), location)
                    yield id_fp
        finally:
            location.save(lineno=lineno)

    def _check_at_start(self):
        if not self._at_start:
            raise TypeError("FPS file is not at the start of the file; cannot search")


    def count_tanimoto_hits_fp(self, query_fp, threshold=0.7):
        """Count the fingerprints which are sufficiently similar to the query fingerprint

        Return the number of fingerprints in the reader which are
        at least *threshold* similar to the query fingerprint *query_fp*.

        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: integer count
        """
        self._check_at_start()
        return fps_search.count_tanimoto_hits_fp(query_fp, self, threshold)

    def count_tanimoto_hits_arena(self, queries, threshold=0.7):
        """Count the fingerprints which are sufficiently similar to each query fingerprint

        Returns a list containing a count for each query fingerprint
        in the *queries* arena. The count is the number of
        fingerprints in the reader which are at least *threshold*
        similar to the query fingerprint.

        The order of results is the same as the order of the queries.
        
        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: list of integer counts, one for each query
        """
        self._check_at_start()
        return fps_search.count_tanimoto_hits_arena(queries, self, threshold)

    def threshold_tanimoto_search_fp(self, query_fp, threshold=0.7):
        """Find the fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this reader which are at least
        *threshold* similar to the query fingerprint *query_fp*.  The
        hits are returned as a :class:`.SearchResult`, in arbitrary
        order.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        self._check_at_start()
        return fps_search.threshold_tanimoto_search_fp(query_fp, self, threshold)

    def threshold_tanimoto_search_arena(self, queries, threshold=0.7):
        """Find the fingerprints which are sufficiently similar to each of the query fingerprints

        For each fingerprint in the *queries* arena, find all of the
        fingerprints in this arena which are at least *threshold*
        similar. The hits are returned as a :class:`.SearchResults`,
        where the hits in each :class:`.SearchResult` is in arbitrary
        order.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResults`
        """
        self._check_at_start()
        return fps_search.threshold_tanimoto_search_arena(queries, self, threshold)

    def knearest_tanimoto_search_fp(self, query_fp, k=3, threshold=0.7):
        """Find the k-nearest fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this reader which are at least
        *threshold* similar to the query fingerprint, and of those, select
        the top *k* hits. The hits are returned as a :class:`.SearchResult`,
        sorted from highest score to lowest.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        self._check_at_start()
        return fps_search.knearest_tanimoto_search_fp(query_fp, self, k, threshold)

    def knearest_tanimoto_search_arena(self, queries, k=3, threshold=0.7):
        """Find the k-nearest fingerprints which are sufficiently similar to each of the query fingerprints

        For each fingerprint in the *queries* arena, find the
        fingerprints in this reader which are at least *threshold*
        similar to the query fingerprint, and of those, select the top
        *k* hits. The hits are returned as a :class:`.SearchResults`,
        where the hits in each :class:`.SearchResult` are sorted by
        similarity score.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResults`
        """
        self._check_at_start()
        return fps_search.knearest_tanimoto_search_arena(queries, self, k, threshold)

def open_output(destination, metadata, compression="", errors="strict", location=None):
    output, close = io.open_compressed_output(destination, compression, "chemfp")
    io.write_fps1_magic(output)
    initial_lineno = 1 + io.write_fps1_header(output, metadata)
    error_handler = io.get_parse_error_handler(errors)

    if location is None:
        location = io.Location.from_destination(destination)
    writer = _fps_writer_gen(output, close, error_handler, location, initial_lineno)
    assert next(writer) == "Ready!" # prime the generator
    
    return FPSWriter(output, writer, metadata, location)

def _fps_writer_gen(output, close, error_handler, location, initial_lineno):
    hexlify = _hexlify
    recno = 0
    output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    def get_lineno():
        return initial_lineno + output_recno
    location.register(get_recno=get_recno, get_output_recno=get_output_recno, get_lineno=get_lineno)
    errno = 0
    try:
        yield_value = "Ready!"
        num_fp_bytes = -1
        while 1:
            id_fp_pairs = (yield yield_value)
            try:
                for id, fp in id_fp_pairs:
                    recno += 1
                    ### Check for illegal characters.
                    if "\t" in id:
                        error_handler.error("Unable to write an identifier containing a tab: %r" % (id,),
                                            location)
                        continue
                    if "\n" in id:
                        error_handler.error("Unable to write an identifier containing a newline: %r" % (id,),
                                            location)
                        continue
                    if not id:
                        error_handler.error("Unable to write a fingerprint with an empty identifier",
                                            location)
                        continue
                    
                    output.write(_hexlify(fp) + "\t" + id + "\n")
                    
                    output_recno += 1
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
            else:
                yield_value = None
                
    finally:
        location.save(recno=get_recno(), output_recno=get_output_recno(), lineno=get_lineno())
        if close is not None:
            try:
                close()
            finally:
                close = None
        
class FPSWriter(FingerprintWriter):
    """Write fingerprints in FPS format.

    This is a subclass of :class:`chemfp.FingerprintWriter`.

    Instances have the following attributes:

    * metadata - a :class:`chemfp.Metadata` instance
    * closed - False when the file is open, else True
    * location - a :class:`chemfp.io.Location` instance

    An FPSWriter is its own context manager, and will close the
    output file on context exit.

    The Location instance supports the "recno", "output_recno",
    and "lineno" properties.
    """
    def __init__(self, output, writer, metadata, location=None):
        "The constructor is not part of the public API"
        self._output = output # Should this be public?
        self._writer = writer
        self.metadata = metadata
        self.location = location
        self.closed = False

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """Close the writer

        This will set self.closed to False.
        """
        if not self.closed:
            self.closed = True
            self._writer.close()


    def write_fingerprint(self, id, fp):
        """Write a single fingerprint record with the given id and fp

        :param string id: the record identifier
        :param bytes fp: the fingerprint
        """
        if self.closed:
            raise ValueError("Cannot write a fingerprint after calling close()")
        err = self._writer.send( [(id, fp)] )
        if err is not None:
            raise err[0], None, err[1]

    def write_fingerprints(self, id_fp_pairs):
        """Write a sequence of fingerprint records

        :param id_fp_pairs: An iterable of (id, fingerprint) pairs.
        """
        if self.closed:
            raise ValueError("Cannot write fingerprints after calling close()")
        err = self._writer.send(id_fp_pairs)
        if err is not None:
            raise err[0], None, err[1]


def parse_date(date, error):
    try:
        return datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S")
    except ValueError as err:
        errmsg = str(err)

    try:
        dt = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError as err:
        pass
    else:
        raise error("The chemfp specification does not allow microseconds in the date: %r"
                        % (date,))

    if "unconverted data remains" in errmsg:
        remainder = date[19:]
        if remainder == "Z" or remainder[:2] in ("+0", "+1", "-0", "-1"):
            raise error(
                "The chemfp specification requires the date be in UTC without a timezone specifier: %r"
                % (date,))
        raise error(
            "Unconverted data remains after the date: %r"
            % (date,))
    
    raise error(
        "The date must be in the form 'YYYY-DD-MMTHH:MM:SS': %r"
        % (date,))
                    

def read_header(f, location):
    #metadata = Metadata()
    num_bits = None
    num_bytes = None
    software = None
    type = None
    sources = []
    date = None
    aromaticity = None

    location.save(lineno=0)

    def error(errmsg):
        location.save(lineno=lineno+1)
        return ParseError(errmsg, location=location)

    lineno = 0
    for block in _read_blocks(f):
        # A block must be non-empty
        start = 0
        while 1:
            c = block[start:start+1]
            if c == b"":
                # End of the block; get the next one
                break
            if c != b'#':
                # End of the header. This block contains the first fingerprint line
                block = block[start:]
                if num_bits is None:
                    if num_bytes is None:
                        # We can figure this out from the fingerprint on the first line
                        err = _chemfp.fps_line_validate(-1, block)
                        if err:
                            if err == -36: # CHEMFP_MISSING_NEWLINE
                                # The block does not end with a newline.
                                # Figure out the correct line number for that line.
                                lineno += block.count(b"\n")
                            raise error(_chemfp.strerror(err))
                        i = block.index(b"\t")
                        # If you don't specify the number of bits then I'll do it for you.
                        num_bits = i * 4
                        num_bytes = i // 2
                    else:
                        num_bits = num_bytes * 8
                else:
                    if num_bytes is None:
                        num_bytes = (num_bits+7) // 8
                    else:
                        # Already validated this case when the values were set.
                        pass

                location.save(lineno=lineno)
                return Metadata(num_bits, num_bytes, type, aromaticity, software,
                                sources, date), block

            start += 1 # Skip the '#'
            end = block.find(b"\n", start)
            if end == -1:
                raise error(_chemfp.strerror(-36)) # CHEMFP_MISSING_NEWLINE
                ## # Only happens when the last line of the file contains
                ## # no newlines. In that case, we're at the last block.
                ## line = block[start:]
                ## start = len(block)
            else:
                line = block[start:end]
                start = end+1

            # Right! We've got a line. Check if it's magic
            # This is the only line which cannot contain a '='
            if lineno == 0:
                if line == b"FPS1" or line == b"#FPS1\r":
                    lineno += 1
                    continue

            if line.startswith(b"x-") or line.startswith(b"X-"):
                # Completely ignore the contents of 'experimental' lines
                continue

            if b"=" not in line:
                raise error("header line must contain an '=': %r" % (line,))
            key, value = line.split(b"=", 1)
            key = key.strip()
            value = value.strip()
            if key == b"num_bits":
                if num_bits is not None:
                    raise error("Only one num_bits header is allowed")
                try:
                    num_bits = int(value)
                    if num_bits <= 0:
                        raise ValueError
                except ValueError:
                    raise error("num_bits header must be a positive integer, not %r" % (value,))

                if num_bytes is not None:
                    n = (num_bits+7)//8
                    if n != num_bytes:
                        raise error(
                            "The num_bits header of %d requires %d bytes, which is "
                            "incompatible with the earlier num_bytes header of %d"
                            %  (num_bits, n, num_bytes))
                                        
            elif key == b"num_bytes":
                if num_bytes is not None:
                    raise error("Only one num_bytes header is allowed")
                try:
                    num_bytes = int(value)
                    if num_bytes <= 0:
                        raise ValueError
                except ValueError:
                    raise error(
                        "num_bytes header must be a positive integer, not %r" % (value,))
                if num_bits is not None:
                    n = (num_bits+7)//8
                    if n != num_bytes:
                        raise error(
                            "The num_bytes header of %d is incompatible with the "
                            "earlier num_bits header of %d, which requires %d bytes"
                            % (num_bytes, num_bits, n))
                            
            elif key == b"software":
                if software is not None:
                    raise error("Only one software header is allowed")
                try:
                    software = value.decode("utf8")
                except UnicodeDecodeError as err:
                    raise error("software header must be utf8 encoded: %r (%s)" % (value, err))
            elif key == b"type":
                if type is not None:
                    raise error("Only one type header is allowed")
                # Should I have an auto-normalization step here which
                # removes excess whitespace?
                #type = normalize_type(value)
                try:
                    type = value.decode("ascii")
                except UnicodeDecodeError as err:
                    raise error("type header must be ASCII encoded: %r (%s)" % (value, err))
            elif key == b"source":
                sources.append(value.decode("utf8", "surrogateescape"))
            elif key == b"date":
                if date is not None:
                    raise error("Only one date header is allowed")
                try:
                    date = value.decode("ascii")
                except UnicodeDecodeError as err:
                    raise error("date header must be ASCII encoded: %r (%s)" % (value, err))
                date = parse_date(date, error)
            elif key == b"aromaticity":
                if aromaticity is not None:
                    raise error("Only one aromaticity header is allowed")
                try:
                    aromaticity = value.decode("ascii")
                except UnicodeDecodeError as err:
                    raise error("aromaticity header must be ASCII encoded: %r (%s)" % (value, err))
            elif key.startswith(b"x-"):
                pass
            else:
                # In the interests of forward compatibility, I ignore unknown headers.
                # XXX Should I give a warning? A log message? How does someone help
                # figure out if there's a typo?
                #print "UNKNOWN", repr(line), repr(key), repr(value)
                #warn(filename, lineno, "Unknown header %r" % (value,))
                pass
            lineno += 1

    # Reached the end of file. No fingerprint lines and nothing left to process.
    if num_bits is None:
        if num_bytes is None:
            num_bits = None
            num_bytes = None
        else:
            num_bits = num_bytes*8
    else:
        if num_bytes is None:
            num_bytes = (num_bits+7)//8
        else:
            # Already verified this case when the values were first set.
            pass

    location.save(lineno=lineno)
    return Metadata(num_bits, num_bytes, type, aromaticity, software,
                    sources, date), None

